From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes.


(* At some points, Diaframe can no longer postpone the choice of the side of a disjunction.
   Instead of hardwiring a particular side-choosing strategy, we make the tactics parametric 
   in such strategies.

   This file defines an interface to define these strategies easily. *)


(* abusing modules as 'namespaces' *)
Module PickSide.
  Inductive PickSide := | Left | Right | Stop.
End PickSide.

Module PickEvent.
  Inductive PickEvent := | NoHintSep | LeftEasier | NoHintOne | LeftModOne.
End PickEvent.

Definition PureChooser := PickEvent.PickEvent → PickSide.PickSide.

Ltac from_pure_chooser pure_fun :=
  fun evt => 
    let the_side := eval cbv in (pure_fun evt) in
    ltac:(fun cont => cont constr:(the_side)).

Definition ChooserAlwaysLeft : PureChooser := λ _, PickSide.Left.
Definition ChooserSafe : PureChooser := 
  λ evt,
    match evt with
    | PickEvent.NoHintSep => PickSide.Left
    | PickEvent.NoHintOne => PickSide.Left
    | _ => PickSide.Stop
    end.
Definition ChooserSafest : PureChooser := λ _, PickSide.Stop.

Ltac chooser_always_left := from_pure_chooser ChooserAlwaysLeft.
Ltac chooser_safe := from_pure_chooser ChooserSafe.
Ltac chooser_safest := from_pure_chooser ChooserSafest.

Ltac chooser_smash evt := fun cont => (cont constr:(PickSide.Left)) + (cont constr:(PickSide.Right)).

(* we would like to have something like the following: 
Ltac chooser_custom side_list :=
  fun evt => 
    lazymatch side_list with
    | @nil PickSide.PickSide => 
      fun cont => cont constr:(PickSide.Stop)
    | cons ?side side_tail =>
      fun cont => 
    end.
  this means a chooser_tac needs to output another chooser_tac.
*)

(* for testing: 
Goal True.
Proof.
  let evt := constr:(PickEvent.NoHintSep) in
  let result := chooser_safest evt in
  let cont := (fun side => lazymatch side with | PickSide.Left => idtac "left ok"; fail | _ => idtac side end) in
  result cont. *)