From diaframe Require Export utils_ltac2.
From iris.proofmode Require Import proofmode environments string_ident.



Ltac2 get_nth_coq_byte (i : int) : constr :=
  let the_inst : instance := 
  (* I dont know another way to get something of type instance! *)
    let k := Constr.Unsafe.kind 'Byte.x00 in
    match k with
    | Constr.Unsafe.Constructor const inst =>
      inst
    | _ => Control.throw Not_found
    end in
  let k := Constr.Unsafe.kind 'Byte.byte in
  match k with
  | Constr.Unsafe.Ind ind _ => 
    let first_cons := Constr.Unsafe.constructor ind i in
    Constr.Unsafe.make (Constr.Unsafe.Constructor first_cons the_inst)
  | _ => Control.throw Not_found 
  end.

Ltac2 char_to_coq_byte (c : char) : constr :=
  get_nth_coq_byte (Char.to_int c).

Ltac2 rec list_to_coq_list (cs : constr list) : constr :=
  match cs with
  | [] => '([]) 
  | c :: cs' =>
    let tail := list_to_coq_list cs' in
    '($c :: $tail)
  end.

Ltac2 rec coq_list_to_list (cs : constr) : constr list :=
  lazy_match! cs with
  | [] => []
  | ?c :: ?cs' => c :: coq_list_to_list cs'
  end.

Ltac2 string_to_coq_string (s : string) : constr :=
  let coq_byte_list : constr list 
    := List.init (String.length s) (fun i => char_to_coq_byte (String.get s i)) in
  let coq_bytes : constr := list_to_coq_list coq_byte_list in
  let raw_result : constr := '(string_of_list_byte $coq_bytes) in
  let result := Std.eval_cbv cbv_flags raw_result in
  result.


(* Coq 8.15.2 does not even have Int.div. T_T
  divide 25 by 2
  find largest 2 power below 25 = 16
  answer = 8 + divide (25 - 16) by 2
  = 8 + divide 9 by 2

  divide 9 by 2
  9 = 8 + 1
  answer = 4 + divide 1 by 2
  12
*)

Ltac2 rec int_div_2 (n : int) : int :=
  if Int.lt n 2 then 0 else
  let rec split_two_power (two_Si : int) (two_i : int) : int * int :=
    let two_SSi := Int.mul two_Si 2 in
    if Int.gt two_SSi n then 
      (two_Si, two_i) 
    else
      split_two_power two_SSi two_Si
  in
  let (to_sub, part) := split_two_power 2 1 in
  Int.add part (int_div_2 (Int.sub n to_sub)).

Ltac2 int_mod_2 (n : int) :=
  Int.sub n (Int.mul 2 (int_div_2 n)).

(* xH = 1
   xO p = 2*p
   xI p = 2*p + 1
*)
Ltac2 rec int_to_coq_positive (i : int) : constr :=
  Control.assert_valid_argument "int_to_coq_positive" (Int.gt i 0);
  if Int.equal i 1 then 'xH else (
  (* i ≥ 2 *)
  let p : int := int_div_2 i in 
  (* TODO: replace with [Int.asr i 1 in] once we drop support for 8.15.2 *)
  let p_constr : constr := int_to_coq_positive p in
  if Int.equal (int_mod_2 i) 0 then
        (* TODO: replace ^ with Int.mod 2 once we drop support for 8.15.2 *)
    (* i = 2 * p *)
    '(xO $p_constr)
  else
    (* i = 2 * p + 1 *)
    '(xI $p_constr)).

Ltac2 rec coq_positive_to_int (p : constr) : int :=
  lazy_match! p with
  | xH => 1
  | xO ?p' => Int.mul 2 (coq_positive_to_int p') 
   (* TODO: replace ^ with [Int.lsl (coq_positive_to_int p') 1] once we drop support for 8.15.2 *)
  | xI ?p' => Int.add (Int.mul 2 (coq_positive_to_int p')) 1
   (* TODO: replace ^ with [Int.lsl (coq_positive_to_int p') 1] once we drop support for 8.15.2 *)
  end.


Ltac2 Type ipm_ident := [
  | NamedIdent (string)
  | AnonIdent (int)
].

Ltac2 ipm_ident_to_constr (ipm_i : ipm_ident) : constr :=
  match ipm_i with
  | NamedIdent s => let coqs := string_to_coq_string s in '(INamed $coqs)
  | AnonIdent i => let coqp := int_to_coq_positive i in '(IAnon $coqp)
  end.

Ltac2 ipm_ident_of_constr (c : constr) : ipm_ident :=
  lazy_match! c with
  | INamed ?j => (NamedIdent (StringToIdent.coq_string_to_string j))
  | IAnon ?p => (AnonIdent (coq_positive_to_int p))
  end.

Ltac2 rec read_ipm_hyp_names_from_env (env : constr) : (ipm_ident * constr) list :=
  lazy_match! env with
  | Enil => []
  | Esnoc ?env' ?coqi ?prop => (ipm_ident_of_constr coqi, prop) :: (read_ipm_hyp_names_from_env env')
  end.

(* We reverse the result since Env are snoc-lists *)
Ltac2 read_ipm_hyp_names (delta : constr) : (ipm_ident * constr) list :=
  lazy_match! delta with
  | Envs ?gamma_i ?gamma_s _ =>
    List.rev (List.append (read_ipm_hyp_names_from_env gamma_s) (read_ipm_hyp_names_from_env gamma_i))
  end.

Ltac2 get_ipm_hyp_names (_ : unit) : (ipm_ident * constr) list :=
  lazy_match! goal with
  | [ |- envs_entails ?delta _ ] =>
    read_ipm_hyp_names delta
  | [ |- _ ] => []
  end.

Ltac2 array_replace (arr : 'a array) (eq : 'a -> 'a -> bool) (old : 'a) (new : 'a) : unit :=
  let replace_el := (fun i el => 
    if (eq el old) then Array.set arr i new else ()
  ) in
  Array.iteri replace_el arr.

Ltac2 string_equal (left : string) (right : string) : bool :=
  let llength := (String.length left) in
  if Int.equal llength (String.length right) then
    let rec cmp (i : int) : bool :=
      if Int.ge i llength then true else
      let cl := Char.to_int (String.get left i) in
      let cr := Char.to_int (String.get right i) in
      if Int.equal cl cr then cmp (Int.add i 1) else false
    in cmp 0
  else false.

Ltac2 ipm_ident_equal (li : ipm_ident) (ri : ipm_ident) : bool :=
  match li with
  | NamedIdent sl =>
    match ri with
    | NamedIdent sr =>
      string_equal sl sr
    | _ => false
    end
  | AnonIdent pl =>
    match ri with
    | AnonIdent pr =>
      Int.equal pl pr
    | _ => false
    end
  end.

Ltac2 option_equal (eq : 'a -> 'a -> bool) : 'a option -> 'a option -> bool :=
  fun ma mb =>
    match ma with
    | Some a =>
      match mb with
      | Some b => eq a b
      | None => false
      end
    | None => false
    end.


Ltac2 ipm_hyp_renamer 
  (register_closure : (unit -> unit) -> unit)
  (remaining_names : ipm_ident list)
  (new_names : constr list) : unit -> unit :=
  fun _ =>
    let current_names := get_ipm_hyp_names () in

    let to_rename := List.filter (fun pr => 
      let (i, _) := pr in
      Bool.neg (List.mem ipm_ident_equal i remaining_names)
    ) current_names in

    let new_names_len := List.length new_names in
    let to_rename_len := List.length to_rename in
    let too_many_names := Int.gt new_names_len to_rename_len in
    let too_little_names := Int.lt new_names_len to_rename_len in

    (* rename fresh IPM hyps as Hs, by repeatedly calling iRename *)
    List.iter (fun pr =>
      let (leftel, j) := pr in
      let (i, _) := leftel in
      let h1 := Ltac1.of_constr (ipm_ident_to_constr i) in
      let h2 := Ltac1.of_constr j in
      ltac1:(H1 H2 |- 
        once (
          iRename H1 into H2 
        + (idtac "Renaming"H1"into"H2"failed! Is"H2"already taken?"))  
      ) h1 h2
    ) (zip to_rename new_names);

    match too_many_names with
    | true =>
      register_closure (fun _ => printf "Too many IPM names were supplied!");
      register_closure (fun _ =>
        printf "Please remove %a" (fun _ ms => List.fold_right Message.concat (fprintf "") ms) 
                                  (List.map (fun c => fprintf "%t, " c) (List.skipn to_rename_len new_names))
      )
    | _ => 
      match too_little_names with
      | true =>
        register_closure (fun _ => printf "Too little IPM names were supplied! Please supply a name for:");
        List.iter (fun (i, p) =>
          let msg := fprintf "  %t : %t" (ipm_ident_to_constr i) p in
          register_closure (fun _ => Message.print msg)
        ) (List.skipn new_names_len to_rename)
      | _ => ()
      end
    end.


Ltac2 rename_ipm_hyps_after_base 
  (prob : NamingProblems)
  (the_tac : (constr -> unit) -> unit) (* tactic to run, requires a tactic to call when deleting ipm hyps *)
  (raw_new_names_list : constr list list) :=
  (* make a snapshot of current IPM hyp names, as a list *)
  let name_store : ipm_ident option array := Array.of_list (List.map (fun (i, _) => Some i) (get_ipm_hyp_names ())) in

  (* make sure that register_delete remove elements from that list *)
  let delete_name_thunk : constr -> unit := 
    (fun (c : constr) =>
        array_replace name_store (option_equal ipm_ident_equal) (Some (ipm_ident_of_constr c)) None
    ) in

  (* run the_tac *)
  the_tac delete_name_thunk;

  initialize_subgoal_problems prob;

  let remaining_names := 
    List.map (fun mi => match mi with Some i => i | _ => Control.throw Not_found end) (
      List.filter (fun i => match i with Some _ => true | _ => false end) (Array.to_list name_store)
    ) in

  let new_names_list := to_list_of_size raw_new_names_list (count_goals ()) []
      (fun i => add_global_problem prob (fun _ => 
        printf "Too little branches of IPM names supplied! Missing %i branches" i 
      ))
      (fun i => add_global_problem prob (fun _ =>
        printf "Too many branches of IPM names supplied! Dropping %i branches" i
      ))
  in

  (* need to use Control.enter to force application of below to single goals. 
      If no goals remain, this is essentially unit. *)
  Control.dispatch (List.mapi (fun i => ipm_hyp_renamer (add_subgoal_problem prob i) remaining_names) new_names_list).

Definition concat_rev (k : list Ascii.ascii) : string := foldl (λ a c, String c a) "" k.

Fixpoint to_string_list_rec (k : string) (res: list string) (cur : list Ascii.ascii) :=
  match k return (list string) with
  | "" => 
      match cur with 
      | [] => res 
      | cur => ((concat_rev cur))::res
      end
  | String a k => 
      if is_space a then 
        match cur with 
        | [] => to_string_list_rec k res cur 
        | cur' => to_string_list_rec k (((concat_rev cur'))::res) [] 
        end
      else
        to_string_list_rec k res (a::cur)
  end.

Definition to_string_list (k : string) := rev (to_string_list_rec k [] []).
Definition to_ident_list (k : string) := map INamed (to_string_list k).

Ltac2 split_ltac1_constr_string (c : Ltac1.t) : constr list :=
  let ltac2_constr := Option.get (Ltac1.to_constr c) in
  coq_list_to_list (Std.eval_cbv cbv_flags '(to_string_list $ltac2_constr)).

Ltac2 rename_multi_ipm_hyps_after_closure 
  (prob : NamingProblems)
  (new_names_list_raw : Ltac1.t)
  (the_tac : (constr -> unit) -> unit) :=
  rename_ipm_hyps_after_base prob the_tac (List.map split_ltac1_constr_string (Option.get (Ltac1.to_list new_names_list_raw))).

Ltac2 rename_ipm_hyps_after_closure 
  (prob : NamingProblems)
  (new_names_raw : Ltac1.t)
  (the_tac : (constr -> unit) -> unit) :=
  rename_ipm_hyps_after_base prob the_tac [split_ltac1_constr_string new_names_raw].

Ltac2 as_ltac1_constr_cps (d : constr -> unit) : Ltac1.t :=
  Ltac1.lambda (fun c : Ltac1.t =>
    d (Option.get (Ltac1.to_constr c)); ltac1val:(idtac)
  ).

Ltac2 ltac1_as_ltac2_cps (d_accepter : Ltac1.t) : (constr -> unit) -> unit :=
  fun (d : constr -> unit) =>
    Ltac1.apply d_accepter [as_ltac1_constr_cps d] Ltac1.run. 


