From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import util_classes solve_defs tele_utils steps.namer_tacs.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file contains 'case 4' of Diaframe's proof search strategy: making progress on atoms behind modalities.
   Lone atoms cause Diaframe to look for an abduction hint.
*)

Section solve_one_rules.

  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma solve_one_fromsep {TT : tele} M P P1 P2 {TT1} P1' {TT2} h M1 M2 P2a :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    SplitLeftModality3 M M1 M2 →
    (TC∀.. (tt1 : TT1), AbsorbModal (∃.. tt2, tele_app P2 (tele_app (tele_app h tt1) tt2))%I
        M2 (tele_app P2a tt1)) → 
    SolveSep M1 P1' P2a
    ⊢ SolveOne M P.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromSepCareful /ExtractDep /AbsorbModal !tforall_forall => 
        HP HP1 [HM HM1 HM2] HP2a.
    rewrite -HM => {HM}.
    apply HM1 => {HM1}.
    apply bi_texist_elim => tt1.
    move: HP1 => /(dep_eval tt1) /tforall_forall HP1.
    rewrite HP2a modality_strong_frame_r.
    apply modality_strong_is_mono => {HM2}.
    apply wand_elim_l', bi_texist_elim => tt2.
    apply wand_intro_l.
    erewrite HP1, HP.
    by rewrite bi_texist_exist -(exist_intro _).
  Qed.

  Lemma solve_one_fromor {TT : tele} M Q L1 L2 :
    (TC∀.. (tt : TT), FromOrCareful (tele_app Q tt) (tele_app L1 tt) (tele_app L2 tt)) →
    ModalityMono M →
    SolveOneFoc (TT := TT) M L1 (∃.. (tt: TT), tele_app L2 tt) ⊢ SolveOne (TT := TT) M Q.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromOrCareful => /tforall_forall HQ HM.
    apply HM, or_elim; apply bi_texist_mono => tt; rewrite -HQ; eauto.
  Qed.

  Lemma solve_one_normalize {TT : tele} M Q Q' :
    (TC∀.. (tt : TT), NormalizedProp (tele_app Q tt) (tele_app Q' tt) true) →
    ModalityMono M →
    SolveOne M Q' ⊢ SolveOne M Q.
  Proof.
    unseal_diaframe; rewrite /TCTForall => /tforall_forall HQ HM.
    apply HM, bi_texist_mono => tt. by rewrite -HQ.
  Qed.

  Lemma solve_one_fromexist_nested {TT1 : tele} {TT2} M P P' :
    (∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    SolveOne (TT := TeleConcatType TT1 TT2) M (tele_curry P') 
    ⊢ SolveOne (TT := TT1) M P.
  Proof.
    unseal_diaframe => /tforall_forall HPP' HM /=.
    apply HM, bi_texist_elim => tt.
    rewrite (tele_arg_concat_inv_eq tt).
    destruct (tele_arg_concat_inv_strong_hd tt) as [tt1 tt2].
    rewrite bi_texist_exist -(exist_intro tt1).
    rewrite -HPP' !bi_texist_exist.
    rewrite -(exist_intro tt2).
    rewrite tele_uncurry_curry_compose.
    by rewrite -tele_curry_uncurry_relation.
  Qed.

  Lemma solve_one_modal M1 M2 M Q Q' :
    Collect1Modal Q M2 Q' →
    CombinedModality M1 M2 M →
    ModalityMono M1 →
    SolveOne (TT := [tele]) M Q' ⊢ SolveOne (TT := [tele]) M1 Q.
    (* this one is necessary if two modalities can only be combined after some evars have been instantiated at a later point
       concretely, this happens if  *)
  Proof.
    unseal_diaframe => /=.
    rewrite /Collect1Modal /CombinedModality => HQ -> HM.
    by apply HM.
  Qed.

  Lemma solve_one_not_introducable {TT'} M Q M1 M2 :
    NotIntroducable M →
    SplitLeftModality3 M M1 M2 →  
    (* TODO: if Q is IntroduceHyp/IntroduceVar, this might cause awkward |={⊤}=> ‖ ∀ -, G goals *)
    SolveSep (TT := [tele]) M1 empty_goal (M2 (∃.. tt, tele_app Q tt)) ⊢
    SolveOne (TT := TT') M Q.
  Proof.
    unseal_diaframe => _ [HM HM' HM''].
    rewrite -HM /=.
    apply HM'.
    rewrite empty_goal_eq left_id //.
  Qed.

  Lemma solve_one_abd {TT} p P M Q R M1 M2 :
    Abduct (TT := TT) p P Q M2 R →
    SplitModality3 M M1 M2 →
    □?p P ∗ M1 R ⊢ SolveOne (TT := TT) M Q.
  Proof.
    unseal_diaframe; rewrite /Abduct => HPR [HM HM1 HM2].
    rewrite -HM modality_strong_frame_r.
    apply HM1 => {HM1} /=.
    by rewrite comm HPR.
  Qed.

  Lemma solve_one_abd_mod_input {TT} p P M Q R :
    AbductModInput TT p P Q M R →
    □?p P ∗ R ⊢ SolveOne (TT := TT) M Q.
  Proof.
    unseal_diaframe => /=. 
    rewrite /AbductModInput /Abduct => HPR //.
  Qed.
End solve_one_rules.


Section coq_tactics.
  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma tac_solve_one_fromsep {TT : tele} Δ M P P1 P2 {TT1} P1' {TT2} h :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    ModalityStrongMono M →
    envs_entails Δ $ SolveSep M P1' (tele_bind (λ tt1, (∃.. tt2, tele_app P2 (tele_app (tele_app h tt1) tt2))%I)) →
    envs_entails Δ $ SolveOne (TT := TT) M P.
  Proof.
    rewrite envs_entails_unseal => HP1P2 HP1P1' HM ->.
    eapply (solve_one_fromsep _ _ _ _ _ _ _ id); try done.
    split; try apply _.
    - move => ? //=.
    - split; move => ?//=.
    - rewrite /TCTForall. apply tforall_forall => tt1.
      rewrite tele_app_bind. rewrite /AbsorbModal /=. reflexivity.
  Qed.

  Lemma tac_solve_one_fromexist_nested {TT1 : tele} Δ {TT2} M P P' :
    (TC∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    envs_entails Δ $ SolveOne (TT := TeleConcatType TT1 TT2) M (tele_curry P') →
    envs_entails Δ $ SolveOne (TT := TT1) M P.
  Proof.
    rewrite envs_entails_unseal => HPP' HM ->.
    by apply solve_one_fromexist_nested.
  Qed.

  Lemma tac_solve_one_ext_abd_input Δ i M {TT} Q R p P :
    FindInExtendedContext Δ (λ p P, AbductModInput TT p P Q M R) i p P →
    envs_entails (envs_option_delete false i p Δ) R →
    envs_entails Δ $ SolveOne (TT := TT) M Q.
  Proof.
    rewrite envs_entails_unseal => HΔiP.
    rewrite (findinextcontext_spec false) => ->.
    apply ficext_satisfies in HΔiP as HPQ.
    rewrite <-solve_one_abd_mod_input => //.
  Qed.

  Lemma tac_solve_one_pure (Δ : envs PROP) {TT : tele} b M P φ :
    (TC∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    IntroducableModality M →
    (* this has the same shape as when encountering a pure sidecondition in SolveSep *)
    (∃.. tt, envs_entails Δ ⌜tele_app φ tt⌝ ∧ TCTrue) →
    (TCIf (TCEq b true) (TCForall Affine (env_spatial Δ)) (TCEq b false)) → 
    envs_entails Δ $ SolveOne (TT := TT) M P.
  Proof.
    rewrite envs_entails_unseal texist_exist => /tforall_forall HPφ HM [tt [Hφ _]].
    unseal_diaframe; rewrite -HM /= bi_texist_exist -(exist_intro tt) -HPφ /=.
    rewrite -(and_idem (of_envs Δ)) {1}Hφ {Hφ}.
    case => [-> HΔ | ->] /=; apply pure_elim_l => Hφ; eauto.
  Qed.

  Lemma tac_solve_one_not_introducable {TT'} Δ M Q M1 M2 :
    NotIntroducable M →
    SplitLeftModality3 M M1 M2 →
    envs_entails Δ $ SolveSep (TT := [tele]) M1 empty_goal (M2 (∃.. tt, tele_app Q tt))%I →
    envs_entails Δ $ SolveOne (TT := TT') M Q.
  Proof.
    rewrite envs_entails_unseal => HM HM' ->.
    apply solve_one_not_introducable => //.
  Qed.

  Lemma tac_solve_one_normalize_or {TT : tele} Δ M Q L1 L2 Q' :
    (TC∀.. (tt : TT), FromOrCareful (tele_app Q tt) (tele_app L1 tt) (tele_app L2 tt)) →
    (TC∀.. (tt : TT), NormalizedProp (tele_app Q tt) (tele_app Q' tt) true) →
    ModalityMono M →
    envs_entails Δ $ SolveOne (TT := TT) M Q' →
    envs_entails Δ $ SolveOne (TT := TT) M Q.
  Proof.
    rewrite envs_entails_unseal => _ HQ HM ->.
    by apply solve_one_normalize.
  Qed.

  Lemma tac_solve_one_fromor {TT : tele} Δ M Q L1 L2 :
    (TC∀.. (tt : TT), FromOrCareful (tele_app Q tt) (tele_app L1 tt) (tele_app L2 tt)) →
    ModalityMono M →
    envs_entails Δ $ SolveOneFoc (TT := TT) M L1 (∃.. (tt: TT), tele_app L2 tt) →
    envs_entails Δ $ SolveOne (TT := TT) M Q.
  Proof.
    rewrite envs_entails_unseal => HQ HM ->.
    by apply solve_one_fromor.
  Qed.

  Lemma tac_solve_one_modal Δ M1 M2 M Q Q' :
    Collect1Modal Q M2 Q' →
    CombinedModality M1 M2 M →
    ModalityMono M1 →
    envs_entails Δ $ SolveOne (TT := [tele]) M Q' →
    envs_entails Δ $ SolveOne (TT := [tele]) M1 Q.
  Proof.
    rewrite envs_entails_unseal => HQ HMs HM1 ->.
    by eapply solve_one_modal.
  Qed.
End coq_tactics.

Ltac solveOneStep namer_tac :=
  first
  [simple notypeclasses refine (tac_solve_one_fromexist_nested _ _ _ _ _ _ _); [ | | tc_solve | tc_solve | simpl]
  |notypeclasses refine (tac_solve_one_fromsep _ _ _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl]
  |notypeclasses refine (tac_solve_one_pure _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl; eauto | tc_solve]
  |notypeclasses refine (tac_solve_one_normalize_or _ _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl]
  |notypeclasses refine (tac_solve_one_fromor _ _ _ _ _ _ _ _); [tc_solve.. | simpl]
  |notypeclasses refine (tac_solve_one_modal _ _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl]
  |notypeclasses refine (tac_solve_one_not_introducable _ _ _ _ _ _ _ _); [tc_solve.. | simpl] 
  |notypeclasses refine (tac_solve_one_ext_abd_input _ _ _ _ _ _ _ _ _); [tc_solve | register_delete namer_tac; cbn; simpl] ].





