From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import env_utils util_classes solve_defs tele_utils steps.namer_tacs.
From iris.bi Require Import bi telescopes.

Import bi.

Section solve_one_foc_rules.
  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma solve_one_foc_fromsep {TT : tele} M Q L1 L2 D TT1 L1' TT2 h:
    (TC∀.. (tt : TT), FromSepCareful (tele_app Q tt) (tele_app L1 tt) (tele_app L2 tt)) →
    ExtractDep L1 TT1 L1' TT2 h →
    ModalityMono M →
    SolveSepFoc (TT := TT1 ) M L1' 
      (tele_bind (λ tt1, ∃..(tt2 : TT2), tele_app L2 (tele_app (tele_app h tt1) tt2))) D ⊢ SolveOneFoc (TT := TT) M Q D.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromSepCareful /ExtractDep !tforall_forall => H HL1 HM.
    apply HM => {HM} /=.
    apply or_mono; last done.
    apply bi_texist_elim => tt1.
    rewrite tele_app_bind.
    apply bi.wand_elim_r', bi_texist_elim => tt2.
    apply bi.wand_intro_l.
    revert HL1 => /(dep_eval tt1) /(dep_eval_tele tt2) ->.
    rewrite bi_texist_exist -(exist_intro) -H //.
  Qed.

  Lemma solve_one_foc_fromexist {TT1 : tele} {TT2} M P P' D :
    (TC∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    SolveOneFoc (TT := TeleConcatType TT1 TT2) M (tele_curry P') D 
    ⊢ SolveOneFoc (TT := TT1) M P D.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromTExist => /tforall_forall HPP' HM.
    apply HM, or_mono; last done.
    apply bi_texist_elim => tt.
    rewrite (tele_arg_concat_inv_eq tt).
    destruct (tele_arg_concat_inv_strong_hd tt) as [tt1 tt2].
    rewrite bi_texist_exist -(exist_intro tt1).
    rewrite -HPP' !bi_texist_exist.
    rewrite -(exist_intro tt2).
    rewrite tele_uncurry_curry_compose.
    by rewrite -tele_curry_uncurry_relation.
  Qed.

  Lemma solve_one_foc_fromor {TT : tele} M Q L1 L2 L1' D b :
    (TC∀.. (tt : TT), FromOrCareful (tele_app Q tt) (tele_app L1 tt) (tele_app L2 tt)) →
    (TC∀.. (tt : TT), NormalizeProp (tele_app L1 tt) (tele_app L1' tt) b ) → 
      (* this tries to make some branches false before investigating deeper, improves iStep behaviour *)
    ModalityMono M →
    SolveOneFoc (TT := TT) M L1' ((∃.. (tt: TT), tele_app L2 tt) ∨ D) ⊢ SolveOneFoc (TT := TT) M Q D.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromOrCareful !tforall_forall => H HL HM.
    apply HM => {HM}.
    rewrite assoc.
    apply or_mono; last done.
    apply or_elim; apply bi_texist_mono => tt; rewrite -H HL; eauto.
  Qed.

  Lemma solve_one_foc_nested_solve_one {TT : tele} M L D :
    SolveOneFoc (TT := TT) M L D ⊢ SolveOneFoc (TT := [tele]) M (SolveOne (TT := TT) id L) D.
  Proof. by unseal_diaframe. Qed.

  Lemma solve_one_foc_nested_solve_sep {TT : tele} M L Q D :
    SolveSepFoc (TT := TT) M L Q D ⊢ SolveOneFoc (TT := [tele]) M (SolveSep (TT := TT) id L Q) D.
  Proof. by unseal_diaframe. Qed.

  Lemma solve_one_foc_focus {TT : tele} M Q D G :
    AsSolveGoalInternal (M $ ∃.. (tt : TT), tele_app Q tt)%I G →
    ModalityMono M →
    G ⊢ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite /AsSolveGoalInternal => -> HM.
    unseal_diaframe. apply HM.
    apply or_intro_l.
  Qed.

  Lemma solve_one_foc_abd_foc_disj_ex {TT} p P M Q R D :
    AbductModInput [tele] p P (FocDisjEx Q D) M R →
    □?p P ∗ R ⊢ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite /AbductModInput /Abduct /= => ->.
    unseal_diaframe.
    by rewrite FocDisjEx_eq /FocDisjEx_def.
  Qed.

  Lemma solve_one_foc_unfocus {TT : tele} M Q D G :
    AsSolveGoalInternal (M D) G →
    ModalityMono M →
    G ⊢ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite /AsSolveGoalInternal => -> HM.
    unseal_diaframe. apply HM.
    apply or_intro_r.
  Qed.

  Lemma solve_one_foc_not_introducable {TT : tele} M M1 M2 Q D :
    NotIntroducable M →
    SplitLeftModality3 M M1 M2 →
    SolveSep (TT := [tele]) M1 (χ) (M2 ((∃.. tt, tele_app Q tt) ∨ D)) ⊢ SolveOneFoc (TT := TT) M Q D.
  Proof.
    unseal_diaframe => /= => _ [HMs HM1 HM2].
    rewrite -HMs. apply HM1. by rewrite empty_goal_eq left_id.
  Qed.
End solve_one_foc_rules.

Section coq_tactics.
  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma tac_solve_one_foc_fromsep {TT : tele} Δ M Q L1 L2 D TT1 L1' TT2 h:
    (TC∀.. (tt : TT), FromSepCareful (tele_app Q tt) (tele_app L1 tt) (tele_app L2 tt)) →
    ExtractDep L1 TT1 L1' TT2 h →
    ModalityMono M →
    envs_entails Δ $ SolveSepFoc (TT := TT1 ) M L1' 
      (tele_bind (λ tt1, ∃..(tt2 : TT2), tele_app L2 (tele_app (tele_app h tt1) tt2))%I) D →
    envs_entails Δ $ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite envs_entails_unseal => HQ HL HM ->.
    by eapply solve_one_foc_fromsep.
  Qed.

  Lemma tac_solve_one_foc_fromexist {TT1 : tele} {TT2} Δ M P P' D :
    (TC∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    envs_entails Δ $ SolveOneFoc (TT := TeleConcatType TT1 TT2) M (tele_curry P') D →
    envs_entails Δ $ SolveOneFoc (TT := TT1) M P D.
  Proof.
    rewrite envs_entails_unseal => HP HM ->.
    by apply solve_one_foc_fromexist.
  Qed.

  Lemma tac_solve_one_foc_pure (Δ : envs PROP) {TT : tele} b M P D φ :
    (∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    IntroducableModality M →
    (TCIf (TCEq b true) (TCForall Affine (env_spatial Δ)) (TCEq b false)) →
    (∃.. tt, tele_app φ tt) →
    envs_entails Δ $ SolveOneFoc (TT := TT) M P D.
  Proof.
    unseal_diaframe; rewrite envs_entails_unseal texist_exist => /tforall_forall HPφ <- + [tt Hφ] /=.
    rewrite bi_texist_exist -(exist_intro tt) -or_intro_l -HPφ pure_True //.
    case => [-> HΔ | ->] /=; eauto.
  Qed.

  Lemma tac_solve_one_foc_fromor {TT : tele} Δ M Q L1 L2 D L1' b :
    (TC∀.. (tt : TT), FromOrCareful (tele_app Q tt) (tele_app L1 tt) (tele_app L2 tt)) →
    (TC∀.. (tt : TT), NormalizeProp (tele_app L1 tt) (tele_app L1' tt) b ) → 
    ModalityMono M →
    envs_entails Δ $ SolveOneFoc (TT := TT) M L1' ((∃.. (tt: TT), tele_app L2 tt) ∨ D) →
    envs_entails Δ $ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite envs_entails_unseal => HQ HL HM ->.
    by eapply solve_one_foc_fromor.
  Qed.

  Lemma tac_solve_one_foc_nested_solve_one {TT : tele} Δ M L D :
    envs_entails Δ $ SolveOneFoc (TT := TT) M L D →
    envs_entails Δ $ SolveOneFoc (TT := [tele]) M (SolveOne (TT := TT) id L) D.
  Proof. rewrite envs_entails_unseal => ->. apply solve_one_foc_nested_solve_one. Qed.

  Lemma tac_solve_one_foc_nested_solve_sep {TT : tele} Δ M L Q D :
    envs_entails Δ $ SolveSepFoc (TT := TT) M L Q D →
    envs_entails Δ $ SolveOneFoc (TT := [tele]) M (SolveSep (TT := TT) id L Q) D.
  Proof. rewrite envs_entails_unseal => ->. apply solve_one_foc_nested_solve_sep. Qed.

  Lemma tac_solve_one_foc_focus {TT : tele} Δ M Q D G :
    AsSolveGoalInternal (M $ ∃.. (tt : TT), tele_app Q tt)%I G →
    ModalityMono M →
    envs_entails Δ G →
    envs_entails Δ $ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite envs_entails_unseal => HQG HM ->.
    apply: solve_one_foc_focus.
  Qed.

  Lemma tac_solve_one_foc_focus_teleO Δ M Q D G :
    AsSolveGoalInternal (M Q)%I G →
    ModalityMono M →
    envs_entails Δ G →
    envs_entails Δ $ SolveOneFoc (TT := [tele]) M Q D.
  Proof.
    rewrite envs_entails_unseal => HQG HM ->.
    apply: solve_one_foc_focus.
  Qed.

  Lemma tac_solve_one_foc_abduct_foc_disj_ex Δ i {TT} p P M Q R D :
    FindInExtendedContext Δ (λ p P, AbductModInput [tele] p P (FocDisjEx Q D) M R) i p P →
    envs_entails (envs_option_delete false i p Δ) R →
    envs_entails Δ $ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite envs_entails_unseal => HΔiP.
    rewrite (findinextcontext_spec false) => ->.
    apply ficext_satisfies in HΔiP as HPQ.
    by rewrite <-solve_one_foc_abd_foc_disj_ex.
  Qed.

  Lemma tac_solve_one_foc_unfocus {TT : tele} Δ M Q D G :
    AsSolveGoalInternal (M D) G →
    ModalityMono M →
    envs_entails Δ G →
    envs_entails Δ $ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite envs_entails_unseal => HG HM ->.
    by apply solve_one_foc_unfocus.
  Qed.

  Lemma tac_solve_one_foc_not_introducable {TT : tele} Δ M M1 M2 Q D :
    NotIntroducable M →
    SplitLeftModality3 M M1 M2 →
    envs_entails Δ (SolveSep (TT := [tele]) M1 (χ)%I (M2 ((∃.. tt, tele_app Q tt) ∨ D)))%I →
    envs_entails Δ $ SolveOneFoc (TT := TT) M Q D.
  Proof.
    rewrite envs_entails_unseal => + + ->.
    apply solve_one_foc_not_introducable.
  Qed.
End coq_tactics.


(* TODO: is this right? *)

From diaframe.steps Require Import pure_solver.

Ltac solvePure :=
  lazymatch goal with
  | |- ex _ => eexists; solvePure
  | |- ?φ ∧ ?R => split; [solvePure | idtac]
  | |- ?φ => solve [trySolvePure]
  end.

From diaframe.steps Require Import disj_chooser_tacs.

Ltac solveOneFocStep namer_tac chooser_tac :=
  first 
  [notypeclasses refine (tac_solve_one_foc_fromsep _ _ _ _ _ _ _ _ _ _ _ _ _ _);
    [tc_solve..| simpl]
  |notypeclasses refine (tac_solve_one_foc_fromor _ _ _ _ _ _ _ _ _ _ _ _);
    [tc_solve..| simpl]
  |notypeclasses refine (tac_solve_one_foc_fromexist _ _ _ _ _ _ _ _);
    [tc_solve..| simpl]
  |notypeclasses refine (tac_solve_one_foc_pure _ _ _ _ _ _ _ _ _ _);
    [tc_solve..| simpl; solvePure ]
  |notypeclasses refine (tac_solve_one_foc_nested_solve_one _ _ _ _ _)
  |notypeclasses refine (tac_solve_one_foc_nested_solve_sep _ _ _ _ _ _)
  | let left_easier := ltac:(fun _ =>
      chooser_tac PickEvent.LeftEasier ltac:(fun side =>
        lazymatch side with
        | PickSide.Left =>
          notypeclasses refine (tac_solve_one_foc_focus_teleO _ _ _ _ _ _ _ _); 
            [tc_solve..| simpl]
        | PickSide.Right =>
           notypeclasses refine (tac_solve_one_foc_unfocus _ _ _ _ _ _ _ _);
             [tc_solve..| simpl]
        | PickSide.Stop => fail
        end
      )
    ) in
    let default_approach := ltac:(fun _ =>
      first 
       [ notypeclasses refine (tac_solve_one_foc_not_introducable _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl]
       | notypeclasses refine (tac_solve_one_foc_abduct_foc_disj_ex _ _ _ _ _ _ _ _ _ _); [tc_solve | simpl]
       | (* only activate NoHintOne if the left hand side is not a modality itself! *)
         lazymatch goal with
         | |- envs_entails _ (SolveOneFoc (TT := [tele]) _ ?L _) =>
           tryif
             assert_succeeds (assert (∃ M P', Collect1Modal L M P') by (do 2 eexists; tc_solve))
           then
             chooser_tac PickEvent.LeftModOne ltac:(fun side =>
              lazymatch side with
              | PickSide.Left => 
                notypeclasses refine (tac_solve_one_foc_focus _ _ _ _ _ _ _ _); 
                  [tc_solve..| simpl]
              | PickSide.Right =>
                notypeclasses refine (tac_solve_one_foc_unfocus _ _ _ _ _ _ _ _);
                  [tc_solve..| simpl]
              | PickSide.Stop => fail
              end
             )
           else
             chooser_tac PickEvent.NoHintOne ltac:(fun side =>
              lazymatch side with
              | PickSide.Left =>
                notypeclasses refine (tac_solve_one_foc_unfocus _ _ _ _ _ _ _ _);
                    [tc_solve..| simpl]
              | _ => fail
              end
             )
         | |- _ => (* nonempty telescope *)
           chooser_tac PickEvent.NoHintOne ltac:(fun side =>
             lazymatch side with
             | PickSide.Left =>
               notypeclasses refine (tac_solve_one_foc_unfocus _ _ _ _ _ _ _ _);
                   [tc_solve..| simpl]
             | _ => fail
             end
            )
         end
       ]
    ) in
    lazymatch goal with
    | |- envs_entails _ (SolveOneFoc (TT := ?TT) _ ?L _) => 
      (* if TT = [tele] and CommitLeftDisjunct L, short-circuit to left_easier *)
      lazymatch TT with
      | [tele] =>
        tryif 
          assert_succeeds (assert (CommitLeftDisjunct L) by tc_solve)
        then
          left_easier ()
        else
          default_approach ()
      | _ =>
        default_approach ()
      end
    end
  ].
