From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import env_utils util_classes solve_defs tele_utils steps.namer_tacs.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file contains 'case 5' of Diaframe's proof search strategy.
*)

Section solve_sep_rules.

  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma solve_sep_mod_weaker M1 M2 {TT} Q R :
    ModWeaker M1 M2 →
    SolveSep (TT := TT) M1 Q R ⊢ SolveSep (TT := TT) M2 Q R.
  Proof. by unseal_diaframe. Qed.

  Lemma solve_sep_id_emp (R : PROP) :
    R ⊢ SolveSep (TT := [tele]) id emp%I R.
  Proof. unseal_diaframe => /=. eauto. Qed.

  Lemma solve_sep_fromsep {TT : tele} M P R P1 P2 {TT1} P1' {TT2} h M1 M2 :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    SplitLeftModality3 M M1 M2 →
    SolveSep M1 P1' $ 
      tele_bind (λ tt1, SolveSep M2 (tele_fun_compose' P2 (tele_app h tt1)) $ tele_fun_compose' R (tele_app h tt1)) 
    ⊢ SolveSep M P R.
  Proof.
    unseal_diaframe; rewrite /TCTForall /FromSepCareful /ExtractDep !tforall_forall => HP HP1 [HM HM1 HM2].
    rewrite -HM => {HM}.
    apply HM1 => {HM1}.
    apply bi_texist_elim => tt1.
    move: HP1 => /(dep_eval tt1) /tforall_forall HP1.
    rewrite tele_app_bind modality_strong_frame_r.
    apply modality_strong_is_mono => {HM2}.
    apply wand_elim_l', bi_texist_elim => tt2.
    apply wand_intro_l.
    rewrite !tele_fun_compose_eq' assoc.
    erewrite HP1, HP.
    by rewrite bi_texist_exist -(exist_intro _).
  Qed.

  Lemma solve_sep_fromexist_nested {TT1 : tele} {TT2} M P P' R :
    (∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    SolveSep (TT := TeleConcatType TT1 TT2) M (tele_curry P') (tele_curry $ tele_dep_bind (λ tt1, tele_bind (λ tt2, tele_app R tt1))) 
    ⊢ SolveSep (TT := TT1) M P R.
  Proof.
    unseal_diaframe => /tforall_forall HPP' HM /=.
    apply HM, bi_texist_elim => tt.
    rewrite (tele_arg_concat_inv_eq tt).
    destruct (tele_arg_concat_inv_strong_hd tt) as [tt1 tt2].
    rewrite bi_texist_exist -(exist_intro tt1).
    rewrite -HPP' !bi_texist_exist.
    rewrite -(exist_intro tt2).
    rewrite tele_uncurry_curry_compose.
    rewrite -tele_curry_uncurry_relation.
    apply sep_mono_r.
    rewrite tele_curry_uncurry_relation.
    rewrite -tele_uncurry_curry_compose.
    rewrite tele_dep_appd_bind.
    by rewrite tele_app_bind.
  Qed.

  Lemma solve_sep_biabd {TTl TTr} M P2 R S T M2 p P1 M1 :
    BiAbd (TTl := TTl) (TTr := TTr) p P1 P2 M2 S T →
    SplitModality3 M M1 M2 →
    □?p P1 ∗ SolveSep (TT := TTl) M1 S $ tele_bind (λ ttl, 
      IntroduceVars (TT := TTr) $ tele_bind (λ ttr, 
        IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R ttr)
    )) ⊢ SolveSep (TT := TTr) M P2 R.
  Proof.
    unseal_diaframe; rewrite /BiAbd => /tforall_forall HP1P2 [HM HM1 HM2].
    rewrite -HM modality_strong_frame_r.
    apply HM1 => {HM1} /=.
    apply wand_elim_l', bi_texist_elim => ttl.
    apply wand_intro_l.
    rewrite tele_app_bind sep_assoc HP1P2 modality_strong_frame_l.
    apply HM2 => {HM2} /=.
    apply wand_elim_l', bi_texist_elim => ttr.
    apply wand_intro_r.
    rewrite (bi_tforall_elim ttr) !tele_app_bind bi_texist_exist -(exist_intro ttr).
    by rewrite -assoc wand_elim_r.
  Qed.

  Lemma solve_sep_biabd_disj {TTl TTr} M P2 R S T M2 p P1 M1 M' Mr D R' :
    SplitLeftModality3 M M' Mr →
    BiAbdDisj (TTl := TTl) (TTr := TTr) p P1 P2 M2 S T D →
    SplitModality3 M' M1 M2 →
    (TC∀.. (ttr : TTr), AbsorbModal (tele_app R ttr) Mr (tele_app R' ttr)) →
    □?p P1 ∗ SolveSep (TT := TTl) M1 S $ tele_bind (λ ttl, 
      match tele_app D ttl with
      | Some2 D' =>
          SolveAnd (
            IntroduceVars (TT := TTr) $ tele_bind $ λ ttr, 
              IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R' ttr)
          ) $ IntroduceHyp D' $ SolveSep (TT := TTr) Mr P2 R
      | None2 => IntroduceVars (TT := TTr) $ tele_bind $ λ ttr, 
              IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R' ttr)
      end
    ) ⊢ SolveSep (TT := TTr) M P2 R.
  Proof.
    unseal_diaframe; rewrite /BiAbdDisj /TCTForall => - [HMM' HM' HMr] /tforall_forall HP1P2 [HM'' HM1 HM2] /tforall_forall HRR'.
    rewrite -HMM' -HM'' modality_strong_frame_r.
    apply HM1 => {HM1} /=.
    apply wand_elim_l', bi_texist_elim => ttl.
    apply wand_intro_l.
    rewrite tele_app_bind sep_assoc HP1P2 modality_strong_frame_l.
    apply HM2 => {HM2} /=.
    destruct (tele_app D ttl) as [D'|].
    - apply wand_elim_l', or_elim, wand_intro_r.
      * apply bi_texist_elim => ttr.
        apply wand_intro_r.
        rewrite and_elim_l (bi_tforall_elim ttr) !tele_app_bind.
        rewrite -assoc wand_elim_r.
        rewrite HRR' modality_strong_frame_r.
        apply HMr.
        rewrite bi_texist_exist -bi.exist_intro comm //.
      * rewrite and_elim_r wand_elim_r //.
    - rewrite right_id. 
      apply wand_elim_l', bi_texist_elim => ttr.
      apply wand_intro_r.
      rewrite (bi_tforall_elim ttr) !tele_app_bind.
      rewrite -assoc wand_elim_r.
      rewrite HRR' modality_strong_frame_r.
      apply HMr.
      rewrite bi_texist_exist -bi.exist_intro comm //.
  Qed.

  Lemma solve_sep_from_or {TT : tele} D L R G M TT1 L' TT2 h G' :
    (TC∀.. (tt : TT), FromOrCareful (tele_app D tt) (tele_app L tt) (tele_app R tt)) →
    ExtractDep L TT1 L' TT2 h →
    (TC∀.. (tt1 : TT1), AbsorbExists (tele_fun_compose' G (tele_app h tt1)) (tele_app G' tt1)) →
    ModalityMono M →
    SolveSepFoc (TT := TT1) M L' G' (∃.. (tt : TT), tele_app R tt ∗ tele_app G tt) ⊢
    SolveSep (TT := TT) M D G.
  Proof.
    unseal_diaframe; rewrite /TCTForall => /tforall_forall HDLR HL /tforall_forall HG HM.
    apply HM, or_elim; last (apply bi_texist_mono => tt; rewrite -HDLR; apply bi.sep_mono_l, or_intro_r).
    apply bi_texist_elim => tt1. rewrite HG.
    apply wand_elim_r', bi_texist_elim => tt2. apply wand_intro_l.
    revert HL. move => /(dep_eval_tele tt1) /(dep_eval_tele tt2) ->.
    rewrite bi_texist_exist -bi.exist_intro. rewrite tele_fun_compose_eq'.
    apply sep_mono_l. rewrite -HDLR. apply or_intro_l.
  Qed.

  Lemma solve_sep_normalize {TT : tele} D G M D' :
    (TC∀.. (tt : TT), NormalizedProp (tele_app D tt) (tele_app D' tt) true) →
    ModalityMono M → (* only used if D is a disjunction *)
    SolveSep (TT := TT) M D' G ⊢
    SolveSep (TT := TT) M D G.
  Proof.
    unseal_diaframe; rewrite /TCTForall !tforall_forall => HD' HM.
    apply HM, bi_texist_mono => tt. apply sep_mono_l. by rewrite -HD'.
  Qed.

End solve_sep_rules.

Section solve_sep_tacs.
  Context {PROP : bi}.
  Implicit Types M : PROP → PROP.

  Lemma tac_solve_sep_emp Δ M R :
    IntroducableModality M →
    envs_entails Δ R →
    envs_entails Δ $ SolveSep (TT := [tele]) M emp%I R.
  Proof.
    rewrite envs_entails_unseal /= => HM ->.
    by rewrite -solve_sep_mod_weaker -solve_sep_id_emp.
  Qed.

  Lemma tac_solve_sep_fromsep {TT : tele} Δ M P R P1 P2 {TT1} P1' {TT2} h :
    (TC∀.. (tt : TT), FromSepCareful (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ExtractDep P1 TT1 P1' TT2 h →
    ModalityStrongMono M →
    envs_entails Δ $ SolveSep M P1' $ 
      tele_bind (λ tt1, SolveSep id (tele_fun_compose' P2 (tele_app h tt1)) $ tele_fun_compose' R (tele_app h tt1)) →
    envs_entails Δ $ SolveSep M P R.
  Proof.
    rewrite envs_entails_unseal => HP1P2 HP1P1' HM ->.
    eapply solve_sep_fromsep; try done. split; try tc_solve. move => ? //=. split => ?? => //=.
  Qed.

  Lemma tac_solve_sep_fromexist_nested {TT1 : tele} Δ {TT2} M P P' R :
    (TC∀.. (tt : TT1), FromTExist (tele_app P tt) (tele_app TT2 tt) (tele_appd_dep P' tt)) →
    ModalityMono M →
    envs_entails Δ $ SolveSep (TT := TeleConcatType TT1 TT2) M (tele_curry P') 
                      (tele_curry $ tele_dep_bind (λ tt1, tele_bind (λ tt2, tele_app R tt1))) →
    envs_entails Δ $ SolveSep (TT := TT1) M P R.
  Proof.
    rewrite envs_entails_unseal => HPP' HM ->.
    by apply solve_sep_fromexist_nested.
  Qed.

  Lemma tac_solve_sep_biabd Δ i M {TTl TTr} P2 R S T M2 p P1 M1 :
    FindInContext Δ (λ p P, BiAbdMod (TTl := TTl) (TTr := TTr) p P P2 M M1 M2 S T) i p P1 →
    envs_entails (envs_delete false i p Δ) 
      (SolveSep (TT := TTl) M1 S (tele_bind $ λ ttl, IntroduceVars (TT := TTr) $ tele_bind (λ ttr, IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R ttr)))) →
    envs_entails Δ $ SolveSep (TT := TTr) M P2 R.
  Proof.
    rewrite envs_entails_unseal => HΔiP HΔ'.
    rewrite (findincontext_spec false) HΔ'.
    apply fic_satisfies in HΔiP as [HPST [HMs HM1 HM2]].
    erewrite solve_sep_biabd => //.
  Qed.

  Lemma tac_solve_sep_ext_biabd Δ mi M {TTl TTr} P2 R S T M2 p P1 M1  :
    FindInExtendedContext Δ (λ p P, BiAbdMod (TTl := TTl) (TTr := TTr) p P P2 M M1 M2 S T) mi p P1 →
    envs_entails (envs_option_delete false mi p Δ)
      (SolveSep (TT := TTl) M1 S (tele_bind $ λ ttl, IntroduceVars (TT := TTr) $ tele_bind (λ ttr, IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R ttr)))) →
    envs_entails Δ $ SolveSep (TT := TTr) M P2 R.
  Proof.
    rewrite envs_entails_unseal => HΔiP HΔ'.
    rewrite (findinextcontext_spec false) // HΔ'.
    apply ficext_satisfies in HΔiP as [HPST [HMs HM1 HM2]].
    erewrite solve_sep_biabd => //.
  Qed.

  Lemma tac_solve_sep_pure_intro Δ {TT : tele} b M P φ Q :
    (TC∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    IntroducableModality M →
    (∃.. tt, tele_app φ tt ∧ envs_entails Δ (tele_app Q tt)) →
    envs_entails Δ $ SolveSep (TT := TT) M P Q.
  Proof.
    rewrite envs_entails_unseal texist_exist => /tforall_forall HPφ HM [tt [Hφ ->]].
    unseal_diaframe; rewrite -HM /= bi_texist_exist -(exist_intro tt) -HPφ /= pure_True //=.
    destruct b => /=;
    iIntros "HQ";
    by iSplitR.
  Qed.

  Lemma tac_solve_sep_pure_intro_2 Δ {TT : tele} b M P φ Q :
    (TC∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    IntroducableModality M →
      (* The user might have run iStep with Δ not simplified.
        In that case, it is better to display the original entailment, 
        so that the user can still solve it. *)
    (∃.. tt, envs_entails Δ ⌜(tele_app φ tt)⌝ ∧ envs_entails Δ (tele_app Q tt)) →
    envs_entails Δ $ SolveSep (TT := TT) M P Q.
  Proof.
    rewrite envs_entails_unseal texist_exist => /tforall_forall HPφ HM [tt [Hφ HQ]].
    rewrite -(and_idem (of_envs Δ)) {1}Hφ {Hφ} HQ.
    apply pure_elim_l => Hφ.
    unseal_diaframe; rewrite -HM /= bi_texist_exist -(exist_intro tt) -HPφ /= pure_True //=.
    destruct b => /=;
    iIntros "HQ";
    by iSplitR.
  Qed.

  (* this one is necessary since |={A,B}=> P ∗ (Q -∗ R) now becomes 
      ‖ |={A,B}=> ‖ P (∗) (Q -∗ R)  since the wand cannot do anything with fupd *)
  Lemma tac_solve_sep_pure_non_intro Δ {TT : tele} b M P φ Q Q' :
    (TC∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    ModalityMono M → 
    (TC∀.. tt : TT, AbsorbModal (tele_app Q tt) M (tele_app Q' tt)) →
    (∃.. tt, tele_app φ tt ∧ envs_entails Δ (tele_app Q' tt)) →
    envs_entails Δ $ SolveSep (TT := TT) M P Q.
  Proof.
    rewrite envs_entails_unseal texist_exist /TCTForall !tforall_forall => HPφ HM HQ [tt [Hφ ->]].
    unseal_diaframe; rewrite HQ; apply HM. 
    rewrite /= bi_texist_exist -(exist_intro tt) -HPφ /= pure_True //=.
    destruct b => /=;
    iIntros "HQ";
    by iSplitR.
  Qed.

  (* this one is necessary since |={A,B}=> P ∗ (Q -∗ R) now becomes 
      ‖ |={A,B}=> ‖ P (∗) (Q -∗ R)  since the wand cannot do anything with fupd *)
  Lemma tac_solve_sep_pure_non_intro_2 Δ {TT : tele} b M P φ Q Q' :
    (TC∀.. tt : TT, FromPure b (tele_app P tt) (tele_app φ tt)) →
    ModalityMono M → 
    (TC∀.. tt : TT, AbsorbModal (tele_app Q tt) M (tele_app Q' tt)) →
    (∃.. tt, envs_entails Δ ⌜BehindModal M (tele_app φ tt)⌝ ∧ envs_entails Δ (tele_app Q' tt)) →
    envs_entails Δ $ SolveSep (TT := TT) M P Q.
  Proof.
    rewrite envs_entails_unseal texist_exist /TCTForall /BehindModal !tforall_forall => HPφ HM HQ [tt [Hφ HQ']].
    rewrite -(and_idem (of_envs Δ)) {1}Hφ {Hφ} HQ'.
    apply pure_elim_l => Hφ.
    unseal_diaframe; rewrite HQ; apply HM. 
    rewrite /= bi_texist_exist -(exist_intro tt) -HPφ /= pure_True //=.
    destruct b => /=;
    iIntros "HQ";
    by iSplitR.
  Qed.

  Lemma tac_solve_sep_ext_biabd_disj Δ mi M {TTl TTr} P2 R S T M2 p P1 M1 M' Mr D R' :
    SplitLeftModality3 M M' Mr →
    FindInExtendedContext Δ (λ p P, BiAbdDisjMod (TTl := TTl) (TTr := TTr) p P P2 M' M1 M2 S T D) mi p P1 →
    (TC∀.. (ttr : TTr), AbsorbModal (tele_app R ttr) Mr (tele_app R' ttr)) →
    envs_entails (envs_option_delete false mi p Δ)
      (SolveSep (TT := TTl) M1 S $ tele_bind (λ ttl, 
        match tele_app D ttl with
        | Some2 D' =>
            SolveAnd (
              IntroduceVars (TT := TTr) $ tele_bind $ λ ttr, 
                IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R' ttr)
            ) $ IntroduceHyp D' $ SolveSep (TT := TTr) Mr P2 R
        | None2 => IntroduceVars (TT := TTr) $ tele_bind $ λ ttr, 
                IntroduceHyp (tele_app (tele_app T ttl) ttr) (tele_app R' ttr)
        end
    )) →
    envs_entails Δ $ SolveSep (TT := TTr) M P2 R.
  Proof.
    rewrite envs_entails_unseal => HMM' HΔiP HR HΔ'.
    rewrite (findinextcontext_spec false) // HΔ'.
    apply ficext_satisfies in HΔiP as [HPST [HMs HM1 HM2]].
    erewrite solve_sep_biabd_disj => // {HΔ'}.
  Qed.

  Lemma tac_solve_sep_from_or Δ {TT : tele} D L R G M TT1 L' TT2 h G' :
    (TC∀.. (tt : TT), FromOrCareful (tele_app D tt) (tele_app L tt) (tele_app R tt)) →
    ExtractDep L TT1 L' TT2 h →
    (TC∀.. (tt1 : TT1), AbsorbExists (tele_fun_compose' G (tele_app h tt1)) (tele_app G' tt1)) →
    ModalityMono M →
    envs_entails Δ (SolveSepFoc (TT := TT1) M L' G' (∃.. (tt : TT), tele_app R tt ∗ tele_app G tt)%I) →
    envs_entails Δ $ SolveSep (TT := TT) M D G.
  Proof.
    rewrite envs_entails_unseal => HD HL HG HM ->.
    by eapply solve_sep_from_or.
  Qed.

  Lemma tac_solve_sep_normalize_or Δ {TT : tele} D L R G M D' :
      (* TODO: the FromOrCareful term is unused, replace with Ltac? *)
    (TC∀.. (tt : TT), FromOrCareful (tele_app D tt) (tele_app L tt) (tele_app R tt)) →
    (TC∀.. (tt : TT), NormalizedProp (tele_app D tt) (tele_app D' tt) true) →
    ModalityMono M →
    envs_entails Δ (SolveSep (TT := TT) M D' G) →
    envs_entails Δ $ SolveSep (TT := TT) M D G.
  Proof.
    rewrite envs_entails_unseal => HD HD' HM ->.
    by apply solve_sep_normalize.
  Qed.

End solve_sep_tacs.

Ltac solveSepStep namer_tac := 
  first
  [simple notypeclasses refine (tac_solve_sep_fromexist_nested _ _ _ _ _ _ _ _); [ | | tc_solve | tc_solve | simpl]
  |notypeclasses refine (tac_solve_sep_fromsep _ _ _ _ _ _ _ _ _ _ _ _);
    [tc_solve.. | simpl] (* by putting pure after fromsep, we ensure that the pure propositions are not ∧ :)*)
(*  |notypeclasses refine (tac_solve_sep_pure_intro_2 _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl]*)
  |notypeclasses refine (tac_solve_sep_pure_non_intro_2 _ _ _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl]
  |notypeclasses refine (tac_solve_sep_normalize_or _ _ _ _ _ _ _ _ _ _ _); [tc_solve.. | simpl] 
  |notypeclasses refine (tac_solve_sep_from_or _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _);
    [tc_solve.. | cbn; simpl]
(*  |notypeclasses refine (tac_solve_sep_ext_biabd _ _ _ _ _ _ _ _ _ _ _ _ _); 
    [tc_solve.. | register_delete namer_tac; simpl] *)
  (* replaces above.  *)
  |notypeclasses refine (tac_solve_sep_ext_biabd_disj _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _);
    [tc_solve.. | register_delete namer_tac; simpl] ].



















