From iris.proofmode Require Import base coq_tactics reduction tactics string_ident.
From diaframe Require Import util_classes util_instances solve_defs tele_utils utils_ltac2.
From diaframe.steps Require Import namer_tacs small_steps solve_instances disj_chooser_tacs ipm_hyp_renamer.
From iris.bi Require Import bi telescopes.

Import bi.

(* This file bundles all files in the steps/ folder, to create the iStep tactic (and friends).
   It also defines notation for Diaframe goals, and adds some support for IPM tactics for these goals.
*)

Export solve_defs.

Lemma from_as_solve_goal {PROP : bi} Δ (P P' : PROP) :
  AsSolveGoalInternal P P' →
  envs_entails Δ P' →
  envs_entails Δ P.
Proof. rewrite /AsSolveGoalInternal => <- //. Qed.

Lemma as_emp_valid_weak {PROP : bi} (φ : Prop) (P : PROP) `{!AsEmpValidWeak φ P} :
  (⊢ P) → φ.
Proof. by apply as_emp_valid_weak. Qed.

Ltac iStartProof2 :=
  first 
  [ iStartProof 
  | (notypeclasses refine (as_emp_valid_weak _ _ _); [tc_solve || fail "iStartProof2: not a BI assertion" | iStartProof; simpl]) ].

Ltac asSolveGoalStep :=
  notypeclasses refine (from_as_solve_goal _ _ _ _ _ );
    [tc_solve
    |simpl].

Ltac iStepDT delete_tac :=
  iStartProof2; asSolveGoalStep; progress (idtac; solveStepsDT delete_tac).

Ltac iStep_ :=
  iStepDT default_delete_tac.

Ltac iStepS := iStep_.

Ltac iStepDT_num n delete_tac :=
  do n (iStepDT delete_tac).

Tactic Notation "iStep" integer(n) :=
  iStepDT_num n default_delete_tac.

Tactic Notation "iStep" := iStep_.

Ltac2 rename_both_after_ltac1_closure
  (b : ReintroductionWarningBehavior) 
  (raw_intro_pats : Ltac1.t) 
  (raw_ipm_names : Ltac1.t)
  (tac1_closure : Ltac1.t) :=
  (* if raw_ipm_names is an empty list, we replace it with a singleton list of '""*)
  let ipm_names :=
    let raw_ipm_name_list := Option.get (Ltac1.to_list raw_ipm_names) in
    if Int.equal (List.length raw_ipm_name_list) 0 then
      Ltac1.of_list [Ltac1.of_constr '""]
    else
      raw_ipm_names
  in
  let probs := no_problems () in
  rename_multi_ipm_hyps_after_closure probs ipm_names (fun d =>
    rename_multi_after_closure probs raw_intro_pats (fun _ => ltac1_as_ltac2_cps tac1_closure d)
  );
  match b with
  | Silent => ()
  | Verbose =>
    print_problems probs
  end.

Ltac2 iStep_num_as_ 
    (num : Ltac1.t) 
    (b : ReintroductionWarningBehavior) 
    (raw_intro_pats : Ltac1.t) 
    (raw_ipm_names : Ltac1.t) : unit :=
  let ltac_closure := (ltac1val:(num |- fun d => iStepDT_num num d) num) in
  (* the control.enter is to get the 'expected' behavior when calling iStep(s) on multiple goals at once *)
  Control.enter (fun _ =>
    rename_both_after_ltac1_closure b raw_intro_pats raw_ipm_names ltac_closure
  ).

Tactic Notation "iStep" integer(n) "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(n xs' hs' |-
    iStep_num_as_ n Verbose xs' hs'
  ) in
  f n xs Hs.

Tactic Notation "iStep" integer(n) "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(n hs' |-
    iStep_num_as_ n Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f n Hs.

Tactic Notation "give_num_to" integer(n) tactic(the_tac) :=
  the_tac n.

Ltac2 get_ltac1_one (_ : unit) :=
  let the_cell : Ltac1.t option ref := { contents := None } in
  let ltac2_lam : Ltac1.t -> Ltac1.t := 
    (fun result => 
      the_cell.(contents) := Some result;
      ltac1val:(idtac))
    in
  ltac1:(d |- give_num_to 1 d) (Ltac1.lambda ltac2_lam);
  Option.get (the_cell.(contents)).

Tactic Notation "iStep" "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(xs' hs' |-
    iStep_num_as_ (get_ltac1_one ()) Verbose xs' hs'
  ) in
  f xs Hs.

Tactic Notation "iStep" "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hs' |-
    iStep_num_as_ (get_ltac1_one ()) Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f Hs.

Ltac stepTacticOrElse solveGoalTac otherGoalTac pureTac :=
  lazymatch goal with
  | |- envs_entails _ ?R =>
    lazymatch R with
    | (SolveSep _ _ _) => solveGoalTac
    | (Transform _) => solveGoalTac
    | (IntroduceVars _) => solveGoalTac
    | (IntroduceHyp _ _) => solveGoalTac
    | (SolveOne _ _) => solveGoalTac
    | (SolveAnd _ _) => solveGoalTac
    | (SolveSepFoc _ _ _ _) => solveGoalTac
    | (SolveOneFoc _ _ _) => solveGoalTac
    | _ => otherGoalTac
    end
  | |- _ => pureTac
  end.

Ltac iStep_base theSolveStep :=
  stepTacticOrElse theSolveStep ltac:(idtac; asSolveGoalStep; theSolveStep) theSolveStep.

Ltac iStepsDT delete_tac :=
  iStartProof2;
  unshelve (progress (repeat (iStep_base ltac:(idtac; solveStepDT delete_tac)))); shelve_unifiable.

Ltac iSteps_ := iStepsDT default_delete_tac.

Ltac iStepsS := iSteps_.

Ltac2 iSteps_as_ (b : ReintroductionWarningBehavior) (raw_intro_pats : Ltac1.t) (raw_ipm_names : Ltac1.t) : unit :=
  Control.enter (fun _ =>
    rename_both_after_ltac1_closure b raw_intro_pats raw_ipm_names ltac1val:(iStepsDT)
  ).

Tactic Notation "iSteps" "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(xs' hs' |-
    iSteps_as_ Verbose xs' hs'
  ) in
  f xs Hs.

Tactic Notation "iSteps" "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hs' |-
    iSteps_as_ Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f Hs.

Tactic Notation "iSteps" := iSteps_.

Ltac iStepDebug :=
  iStartProof2; asSolveGoalStep.

Ltac iStepSafeDT delete_tac := 
  iStartProof2; asSolveGoalStep; progress (solveStepsNT ltac:(default_namer delete_tac) chooser_safe).

Ltac iStepSafe_ := iStepSafeDT default_delete_tac.

Ltac2 iStepSafe_as_ (b : ReintroductionWarningBehavior) (raw_intro_pats : Ltac1.t) (raw_ipm_names : Ltac1.t) : unit :=
  Control.enter (fun _ =>
    rename_both_after_ltac1_closure b raw_intro_pats raw_ipm_names ltac1val:(iStepSafeDT)
  ).

Tactic Notation "iStepSafe" "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(xs' hs' |-
    iStepSafe_as_ Verbose xs' hs'
  ) in
  f xs Hs.

Tactic Notation "iStepSafe" "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hs' |-
    iStepSafe_as_ Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f Hs.

Tactic Notation "iStepSafe" := iStepSafe_.


Ltac iStepsSafeDT delete_tac := 
  iStartProof2;
  unshelve (progress (repeat (iStep_base ltac:(idtac; solveStepNT ltac:(default_namer delete_tac) chooser_safe)))); shelve_unifiable.

Ltac iStepsSafe_ := iStepsSafeDT default_delete_tac.

Ltac2 iStepsSafe_as_ (b : ReintroductionWarningBehavior) (raw_intro_pats : Ltac1.t) (raw_ipm_names : Ltac1.t) : unit :=
  Control.enter (fun _ =>
    rename_both_after_ltac1_closure b raw_intro_pats raw_ipm_names ltac1val:(iStepsSafeDT)
  ).

Tactic Notation "iStepsSafe" "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(xs' hs' |-
    iStepsSafe_as_ Verbose xs' hs'
  ) in
  f xs Hs.

Tactic Notation "iStepsSafe" "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hs' |-
    iStepsSafe_as_ Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f Hs.

Tactic Notation "iStepsSafe" := iStepsSafe_.


Ltac iStepSafestDT delete_tac := 
  iStartProof2; asSolveGoalStep; progress (solveStepsNT ltac:(default_namer delete_tac) chooser_safest).

Ltac iStepSafest_ := iStepSafestDT default_delete_tac.

Ltac2 iStepSafest_as_ (b : ReintroductionWarningBehavior) (raw_intro_pats : Ltac1.t) (raw_ipm_names : Ltac1.t) : unit :=
  Control.enter (fun _ =>
    rename_both_after_ltac1_closure b raw_intro_pats raw_ipm_names ltac1val:(iStepSafestDT)
  ).

Tactic Notation "iStepSafest" "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(xs' hs' |-
    iStepSafest_as_ Verbose xs' hs'
  ) in
  f xs Hs.

Tactic Notation "iStepSafest" "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hs' |-
    iStepSafe_as_ Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f Hs.

Tactic Notation "iStepSafest" := iStepSafest_.


Ltac iStepsSafestDT delete_tac := 
  iStartProof2;
  unshelve (progress (repeat (iStep_base ltac:(idtac; solveStepNT ltac:(default_namer delete_tac) chooser_safest)))); shelve_unifiable.

Ltac iStepsSafest_ := iStepsSafestDT default_delete_tac.

Ltac2 iStepsSafest_as_ (b : ReintroductionWarningBehavior) (raw_intro_pats : Ltac1.t) (raw_ipm_names : Ltac1.t) : unit :=
  Control.enter (fun _ =>
    rename_both_after_ltac1_closure b raw_intro_pats raw_ipm_names ltac1val:(iStepsSafestDT)
  ).

Tactic Notation "iStepsSafest" "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(xs' hs' |-
    iStepsSafest_as_ Verbose xs' hs'
  ) in
  f xs Hs.

Tactic Notation "iStepsSafest" "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hs' |-
    iStepsSafest_as_ Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f Hs.

Tactic Notation "iStepsSafest" := iStepsSafest_.


(* constructing iSmash takes a bit more effort *)

Ltac smashSolveStep := solveStepNT ltac:(anon_namer default_delete_tac) chooser_smash.

Ltac smashOtherGoalStep :=
  lazymatch goal with
  | |- _ ∧ _ => split; [solve [maybeIPureIntro; pure_solver.trySolvePure]|]
      (* overrides solveSteps shelving of unsolvable pure goals that contain no evars  *)
  | |- _ => solveStepNT ltac:(anon_namer default_delete_tac) chooser_smash (* pure thingies *)
  end.

Ltac smashStep := stepTacticOrElse ltac:(idtac; smashSolveStep) ltac:(idtac; asSolveGoalStep) ltac:(idtac; smashOtherGoalStep).

Ltac smashSteps := 
  smashStep; once solve [unshelve smashSteps; apply inhabitant].

Ltac iSmash := iStartProof2; smashSteps.


Section iris_tactic_compatibility.
  Context {PROP : bi}.

  (* Makes iExist work on SolveSep goals *)
  Global Instance solve_sep_into_exist `{TTf : A → tele} M (P Q : TeleS TTf -t> PROP) :
    ModalityStrongMono M →
    FromExist (SolveSep (TT := TeleS TTf) M P Q) (λ a, SolveSep (TT := TTf a) M (P a) (Q a)).
  Proof.
    unseal_diaframe; rewrite /FromExist => HM.
    apply exist_elim => a.
    apply HM.
    rewrite !bi_texist_exist.
    apply exist_elim => tt.
    by rewrite -(exist_intro $ TargS a tt) /=.
  Qed.

  (* Makes iIntros "H" work on IntroduceHyp *)
  Global Instance introduce_hyp_into_wand (Q R : PROP) :
    FromWand (IntroduceHyp Q R) Q R. (* also makes AsSolveGoal work *)
  Proof. unseal_diaframe. tc_solve. Qed.

  (* Makes iIntros (x) work on IntroduceVars *)
  Global Instance introduce_vars_into_forall {A : Type} {b : A → tele} (Q : TeleS b -t> PROP) :
    FromForall (IntroduceVars Q) (λ a : A, IntroduceVars (TT := b a) (Q a)) (λ x, x).
  Proof. unseal_diaframe; tc_solve. Qed.

  (* Makes iMod work on SolveSep *)
  Global Instance solve_sep_elim_modal p (P : PROP) P' M M1 M2 {TT} R S T :
    IntoModal p P M1 P' →
    TCEq T (SolveSep M (TT := TT) R S) →
    SplitModality3 M M1 M2 →
    ElimModal True p false P P' T (SolveSep M2 (TT := TT) R S).
  Proof.
    unseal_diaframe; rewrite /IntoModal /ElimModal /= => -> -> [<- HM1 HM2] _.
    rewrite modality_strong_frame_l.
    apply HM1.
    by rewrite wand_elim_r.
  Qed.

  (* Makes iSplitL/iSplitR work on SolveSep *)
  Global Instance solvesep_fromsep (L G : PROP) M :
    ModalityStrongMono M →
    FromSep (SolveSep (TT := TeleO) M L G) (M L) G | 999. 
    (* note that G will have its own modality by construction *)
    (* TODO: this is no longer true! *)
  Proof.
    rewrite /FromSep. iSteps. unseal_diaframe => /=.
    rewrite -modality_strong_frame_l. eauto with iFrame.
  Qed.

  (* Makes iLeft/iRight work on SolveSep *)
  Lemma solve_sep_first_from_or {TT : tele} (M : PROP → PROP) P P1 P2 R :
    (TC∀.. tt : TT, FromOr (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ModalityMono M →
    FromOr (SolveSep (TT := TT) M P R) (SolveSep (TT := TT) M P1 R) (SolveSep (TT := TT) M P2 R).
  Proof.
    unseal_diaframe; rewrite /FromOr => /tforall_forall HPs HM.
    apply bi.or_elim; apply HM; apply bi_texist_mono => tt.
    all: apply sep_mono_l; rewrite -HPs.
    - rewrite -bi.or_intro_l //.
    - rewrite -bi.or_intro_r //.
  Qed.

  (* Makes iLeft/iRight work on SolveOne *)
  Lemma solve_one_from_or {TT : tele} (M : PROP → PROP) P P1 P2 :
    (TC∀.. tt : TT, FromOr (tele_app P tt) (tele_app P1 tt) (tele_app P2 tt)) →
    ModalityMono M →
    FromOr (SolveOne (TT := TT) M P) (SolveOne (TT := TT) M P1) (SolveOne (TT := TT) M P2).
  Proof.
    unseal_diaframe; rewrite /FromOr => /tforall_forall HPs HM.
    apply bi.or_elim; apply HM; apply bi_texist_mono => tt.
    all: rewrite -HPs.
    - rewrite -bi.or_intro_l //.
    - rewrite -bi.or_intro_r //.
  Qed.

  Lemma solve_sep_foc_from_or {TT : tele} (M : PROP → PROP) L Q D :
    ModalityMono M →
    FromOr (SolveSepFoc (TT := TT) M L Q D) (SolveSep M L Q) (M D).
  Proof. unseal_diaframe; rewrite /FromOr => HM. apply or_elim; apply HM; eauto. Qed.

  Lemma solve_one_foc_from_or {TT : tele} (M : PROP → PROP) Q D :
    ModalityMono M →
    FromOr (SolveOneFoc (TT := TT) M Q D) (SolveOne M Q) (M D).
  Proof. unseal_diaframe; rewrite /FromOr => HM. apply or_elim; apply HM; eauto. Qed. 

  (* Makes iInv work on SolveSep *)
  Global Instance solvesep_elimacc `{!BiFUpd PROP} E E1 E2 {X : Type} α β mγ {TT} (R : TT -t> PROP) Q :
    ElimAcc (X := X) True (fupd E1 E2) (fupd E2 E1) α β mγ (SolveSep (TT := TT) (fupd E1 E) R Q) 
      (λ x, SolveSep (TT := [tele]) (fupd E2 E2) (β x) (IntroduceHyp (default emp (mγ x))%I (SolveSep (TT := TT) (fupd E1 E) R Q))).
  Proof.
    unseal_diaframe; rewrite /ElimAcc /accessor => _ /=.
    iIntros "Hx1 Hx2".
    iMod "Hx2" as (x) "[Hα Hβ]".
    iSpecialize ("Hx1" with "Hα").
    iMod "Hx1" as "[Hβ' Hr]".
    iSpecialize ("Hβ" with "Hβ'").
    iMod "Hβ".
    iSpecialize ("Hr" with "Hβ") => //.
  Qed.

  (* Makes iMod work on SolveSep (TODO: wait what does [solve_sep_elim_modal] then accomplish?) *) 
  Global Instance solve_sep_elim_modal_fupd `{!BiFUpd PROP}  (P : PROP) E1 E2 E3 p {TT} A G: 
    ElimModal True p false (|={E1, E2}=> P)%I P (SolveSep (TT := TT) (fupd E1 E3) A G) (SolveSep (TT := TT) (fupd E2 E3) A G).
  Proof.
    rewrite /ElimModal /= bi.intuitionistically_if_elim => _; iStep as "HMP HG".
    unseal_diaframe. iMod "HMP".
    by iMod ("HG" with "HMP") as "$".
  Qed.

  (* Makes iMod work on SolveSep *) 
  Global Instance solve_sep_elim_modal_bupd `{BiBUpdFUpd PROP} (P : PROP) E1 E2 p {TT} A G: 
    ElimModal True p false (|==> P)%I P (SolveSep (TT := TT) (fupd E1 E2) A G) (SolveSep (TT := TT) (fupd E1 E2) A G).
  Proof.
    rewrite /ElimModal /= bi.intuitionistically_if_elim => _; iStep as "HMP HG".
    unseal_diaframe. iMod "HMP".
    by iMod ("HG" with "HMP").
  Qed.

  (* Used for the iDecompose tactic, which reruns a hypothesis through Diaframe's introduction steps *)
  Lemma tac_ireintro i (Δ : envs PROP) p P G :
    envs_lookup i Δ = Some (p, P) →
    envs_entails (envs_delete true i p Δ) (IntroduceHyp (□?p P) G) →
    envs_entails Δ G.
  Proof.
    unseal_diaframe => HΔi.
    rewrite envs_entails_unseal => HPG.
    rewrite envs_lookup_sound // HPG.
    apply bi.wand_elim_r.
  Qed.

End iris_tactic_compatibility.

Ltac iDecompose_base delete_tac hyp_name :=
  eapply (tac_ireintro (INamed hyp_name)); [reflexivity | simpl];
  repeat (introStepDT delete_tac).

Ltac2 iDecompose_base_ 
  (raw_hyp_name : Ltac1.t) 
  (b : ReintroductionWarningBehavior) 
  (raw_intro_pats : Ltac1.t) 
  (raw_ipm_names : Ltac1.t) : unit :=
  rename_both_after_ltac1_closure b raw_intro_pats raw_ipm_names (ltac1val:(hyp |- fun d => iDecompose_base d hyp) raw_hyp_name).

Tactic Notation "iDecompose" constr(hyp_name) "as" "(" simple_intropattern_list_list_sep(xs, "|") ")" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hyp xs' hs' |-
    iDecompose_base_ hyp Verbose xs' hs'
  ) in
  f hyp_name xs Hs.

Tactic Notation "iDecompose" constr(hyp_name) "as" constr_list_sep(Hs, "|") :=
  let f := ltac2:(hyp hs' |-
    iDecompose_base_ hyp Verbose (Ltac1.of_list [Ltac1.of_list []]) hs'
  ) in
  f hyp_name Hs.

Tactic Notation "iDecompose" constr(hyp_name) := iDecompose_base default_delete_tac hyp_name.

Global Hint Extern 4 (FromOr (SolveSep _ _ _) _ _) =>
    eapply solve_sep_first_from_or; solve [tc_solve] : typeclass_instances.

Global Hint Extern 4 (FromOr (SolveOne _ _) _ _) =>
    eapply solve_one_from_or; solve [tc_solve] : typeclass_instances.

Global Hint Extern 4 (FromOr (SolveSepFoc _ _ _ _) _ _) =>
    eapply solve_sep_foc_from_or; solve [tc_solve] : typeclass_instances.

Global Hint Extern 4 (FromOr (SolveOneFoc _ _ _) _ _) =>
    eapply solve_one_foc_from_or; solve [tc_solve] : typeclass_instances.


(* TODO: Add simpler notation for the common case where M := id *)

Notation "'‖' M '‖' ∃ x1 .. xn '‖' L '(∗)' R" :=
  (SolveSep 
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    M
    (λ x1, .. (λ xn, L%I) .. )
    (λ x1, .. (λ xn, R%I) .. )
  )
  (at level 20, M at level 9, x1 binder, xn binder, L, R at level 200, only printing, format
  "‖  M  ‖  ∃  x1 .. xn  ‖  L  (∗)  R"
).

Notation "'‖' M '‖' L '(∗)' R" :=
  (SolveSep 
    (TT := TeleO)
    M
    L
    R
  )
  (at level 20, M at level 9, L, R at level 200, only printing, format
  "‖  M  ‖  L  (∗)  R"
).


Notation "'‖' M '‖' ∃ x1 .. xn '‖' L" :=
  (SolveOne 
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    M
    (λ x1, .. (λ xn, L%I) .. )
  )
  (at level 20, M at level 9, x1 binder, xn binder, L at level 200, only printing, format
  "‖  M  ‖  ∃  x1 .. xn  ‖  L"
).

Notation "'‖' M '‖' L" :=
  (SolveOne 
    (TT := TeleO)
    M
    L
  )
  (at level 20, M at level 9, L at level 200, only printing, format
  "‖  M  ‖  L"
).


Notation "'‖' ∀ x1 .. xn ',' G" :=
  (IntroduceVars
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    (λ x1, .. (λ xn, G%I) .. )
  )
  (at level 20, x1 binder, xn binder, G at level 200, format
  "‖  ∀  x1 .. xn  ,  G"
).

Notation "'‖' ∀ '-' ',' G" :=
  (IntroduceVars
    (TT := TeleO)
    G
  )
  (at level 20, G at level 200, format
  "‖  ∀  -  ,  G"
).


Notation "'‖' L (-∗) R" :=
  (IntroduceHyp L R)
  (at level 20, L, R at level 200, format
  "‖  L  (-∗)  R"
).


Notation "'‖' M '‖' ∃ x1 .. xn '‖' L '(∗)' R '(∨)' G" :=
  (SolveSepFoc 
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    M
    (λ x1, .. (λ xn, L%I) .. )
    (λ x1, .. (λ xn, R%I) .. )
    G
  )
  (at level 20, M at level 9, x1 binder, xn binder, L, R, G at level 200, only printing, format
  "‖  M  ‖ '//' '[   '   ∃  x1 .. xn  ‖  L  (∗)  R ']' '//' (∨) '//' '[   '   G ']'"
).

Notation "'‖' M '‖' L '(∗)' R '(∨)' G" :=
  (SolveSepFoc 
    (TT := TeleO)
    M
    L
    R
    G
  )
  (at level 20, M at level 9, L, R, G at level 200, only printing, format
  "‖  M  ‖ '//' '[   '   L  (∗)  R ']' '//' (∨) '//' '[   '   G ']'"
).



Notation "'‖' M '‖' ∃ x1 .. xn '‖' L '(∨)' G" :=
  (SolveOneFoc 
    (TT := TeleS (λ x1, .. (TeleS (λ xn, TeleO)) ..))
    M
    (λ x1, .. (λ xn, L%I) .. )
    G
  )
  (at level 20, M at level 9, x1 binder, xn binder, L, G at level 200, only printing, format
  "‖  M  ‖ '//' '[   '   ∃  x1 .. xn  ‖  L ']' '//' '(∨)' '//' '[   '   G ']'"
).

Notation "'‖' M '‖' L '(∨)' G" :=
  (SolveOneFoc
    (TT := TeleO)
    M
    L
    G
  )
  (at level 20, M at level 9, L, G at level 200, only printing, format
  "‖  M  ‖ '//' '[   '   L ']' '//' (∨) '//' '[   '   G ']'"
).

















