From diaframe.reloc Require Import proof_automation symb_exec_logatom backtracking reloc_auto_lob.
From diaframe.heap_lang Require Import atomic_specs wp_auto_lob.
From reloc.lib Require Import lock.

From diaframe.lib Require Import own_hints.
From iris.algebra Require Import auth excl numbers.


Definition inc : val := 
  rec: "f" "l" :=
    let: "n" := ! "l" in
    if: CAS "l" "n" ("n" + #1) then
      "n"
    else
      "f" "l".

Section inc_spec.
Context `{!heapGS Σ}.

Instance inc_spec (l : loc) :
  SPEC (z : Z), << l ↦ #z >> inc #l << RET #z; l ↦ #(z + 1) >>.
Proof. iSteps. Qed.

End inc_spec.

Definition fg_incrementer : val :=
  λ: <>, 
    let: "l" := ref #1 in
    (rec: "f" <> :=
      let: "n" := ! "l" in
      if: CAS "l" "n" ("n" + #1) then
        "n"
      else
        "f" #()).

Definition cg_incrementer : val :=
  λ: <>,
    let: "l" := ref #1 in
    let: "lk" := newlock #() in
    (λ: <>,
      acquire "lk";;
      let: "n" := ! "l" in 
      "l" <- "n" + #1 ;;
      release "lk";;
      "n").

Section refinement.
Context `{!relocG Σ}.

Lemma fg_cg_incrementer_refinement :
  ⊢ REL fg_incrementer << cg_incrementer : () → () → lrel_int.
Proof.
  iSteps.
  iAssert (|={⊤}=> inv (nroot.@"incr") 
      (∃ (n : nat), x ↦ #n ∗ x0 ↦ₛ #n ∗ is_locked_r x1 false))%I 
      with "[-]" as ">#Hinv"; first iSteps.
  iSteps.
Qed.
End refinement.

(* above establishes logical refinment, contextual refinement follows: *)
Lemma fg_cg_incrementer_ctx_refinement :
  ∅ ⊨ fg_incrementer ≤ctx≤ cg_incrementer : () → () → TNat.
Proof.
  pose (Σ := #[relocΣ]).
  eapply (refines_sound Σ).
  iIntros (? Δ).
  iApply fg_cg_incrementer_refinement.
Qed.


From iris.program_logic Require Import atomic.
From iris.heap_lang Require Import proofmode.

Definition incr_stateUR := authUR $ optionUR $ exclR natO.
Class incrStateG Σ := IncrStateG {
  incrState_inG :> inG Σ incr_stateUR
}.
Definition incr_stateΣ := #[ GFunctor incr_stateUR].

Global Instance subG_incrStateΣ {Σ} : subG incr_stateΣ Σ → incrStateG Σ.
Proof. solve_inG. Qed.

Section logatom_spec_incrementer.
Context `{!heapGS Σ, !incrStateG Σ}.

Let Nincr := nroot.@"fg_incr".

Instance fg_incrementer_spec :
  SPEC {{ True }} fg_incrementer #() {{ (v : val) γ, RET v;
    own γ (● Excl' 1) ∗ 
  (*  □ <<< ∀∀ n : nat, own γ (● Excl' n) >>>
          v #() @ ↑Nincr
        <<< own γ (● Excl' (S n)), RET #n >>>  
    atomic_wp notation is broken, cf https://gitlab.mpi-sws.org/iris/iris/-/issues/473 *)
    □ atomic_wp (TA := [tele_pair nat]) (TB := [tele]) (v #())%E (↑Nincr)
      (tele_app $ (λ n, own γ (● Excl' n)))
      (tele_app $ (λ n, tele_app (TT := TeleO) (own γ (● Excl' (S n)))))
      (tele_app $ (λ n, tele_app (TT := TeleO) (#n)))
    }}.
Proof.
  iSteps.
  iAssert (|={⊤}=> inv Nincr (∃ (n : nat), x ↦ #n ∗ own x0 (◯ Excl' n)))%I with "[-]" as ">#Hinv".
  { iSteps. }
  iSteps.
  iMod "H3". iReIntro "H3".
  iAssert (|={∅, ⊤ ∖ ↑Nincr}=> emp -∗ own x5 (◯ Excl' (S x2)) ∗ x4 #x2)%I with "[-]" as ">H5".
  { iPoseProof (diaframe_close_right with "H6") as "H6". iSteps. }
  replace (Z.to_nat (x2 + 1)) with (S x2) by lia.
  iSteps.
Qed.
End logatom_spec_incrementer.

