From diaframe Require Import proofmode_base.
From iris.bi Require Import bi.
From iris.proofmode Require Import proofmode.

Section test.
Context `{!BiAffine PROP}.
Implicit Types A B C D : PROP.

Context (P : nat → PROP).

Fixpoint P_or (n : nat) : PROP :=
  match n with
  | O => P 0
  | S n => P_or n ∨ P (S n)
  end.

Fixpoint P_sep (n : nat) : PROP :=
  match n with
  | O => P 0
  | S n => P_sep n ∗ P (S n)
  end.

Definition P_prob (n : nat) : PROP := P_sep (S n) -∗ (P_or (S n) ∗ P_sep n) ∧ (P_or (S n) ∗ P_sep n).

Lemma P_prob_taut : emp ⊢ ∀ n, P_prob n.
Proof. iSmash. Qed.

(* The problem was that the ∧ caused backtracking across the and branches, dramatically worsening performance.
   This should no longer be the case: time (iStep; iSmash) and time (iSmash) should give comparable timings. *)
Let n := 4%nat.

Lemma test_and_4 : P_sep (S n) ⊢ (P_or (S n) ∗ P_sep n) ∧ (P_or (S n) ∗ P_sep n).
Proof. 
  simpl. iStep.
  time (iStep; iSmash). (* solving it twice, in parallel *)
  Undo 1.
  time iSmash. (* solving it twice, should also be in parallel *)
Qed.

Lemma test_fail_4 : P_sep (S n) ⊢ (P_or (S n) ∗ True) ∧ (P_or (S n) ∗ P_sep n ∗ P (S (S n))).
Proof. 
  simpl. iStep.
  Time (iStep; first iSmash).
  Time Fail iSmash. (* try proving second goal, failure time should be comparable to total failure time *)
  Undo 2.
  Time Fail (iStep; iSmash). (* trying to solve in parallel. first goal is provable in different ways, 
        second goal unprovable so causes failure. *)
  Time Fail iSmash. (* solving it twice, should also be in parallel *)
Abort.

End test.