From iris Require Import bi.bi.
From iris.proofmode Require Import proofmode.
From simuliris.simulation Require Import slsls lifting.
From simuliris.simulang Require Import lang notation tactics
  proofmode log_rel_structural wf behavior.
From simuliris.simulang.na_inv Require Export inv.
From simuliris.simulang.na_inv Require Import readonly_refl adequacy.
From iris.prelude Require Import options.

From diaframe.simuliris.na Require Import proof_automation.

Import bi.

(** * Examples for exploiting UB of data-races. *)

Section data_race.
  Context `{naGS Σ}.

  (** This optimization replaces multiple loads with a single load. *)
  Definition remove_load_s : expr :=
    "l" <- !"l" + !"l";;
    !"l".

  Definition remove_load_t : expr :=
    let: "v" := !"l" in
    let: "r" := "v" + "v" in
    "l" <- "r";;
    "r".

  Lemma remove_load_sim:
    ⊢ log_rel remove_load_t remove_load_s.
  Proof.
    log_rel. iSteps.
    rewrite delete_insert //.
    iSteps.
  Qed.

  (** This optimization hoists the read of "x" and "y" out of the loop. *)
  Definition hoist_load_both_s : expr :=
     let: "i" := ref #0  in
     let: "sum" := ref (!"y") in
     while: !"i" ≠ ! "x" do
       "i" <- !"i" + #1;;
       "sum" <- !"sum" + !"y"
     od;;
     let: "res" := !"sum" in
     Free "sum";;
     Free "i";;
     "res".

  Definition hoist_load_both_t : expr :=
     let: "n" := !"x" in
     let: "m" := !"y" in
     let: "i" := ref #0  in
     let: "sum" := ref "m" in
     while: !"i" ≠ "n" do
       "i" <- !"i" + #1;;
       "sum" <- !"sum" + "m"
     od;;
     let: "res" := !"sum" in
     Free "sum";;
     Free "i";;
     "res".


(*  Lemma int_is_unboxed (z : Z) : val_is_unboxed #z.
  Proof. done. Qed.*)
  Lemma hoist_load_both_sim:
    ⊢ log_rel hoist_load_both_t hoist_load_both_s.
  Proof.
    log_rel.
    iSteps.
    destruct (if x2 is #(LitLoc _) then true else false) eqn:Heq. 2: {
      source_while.
      source_bind (! _)%E. iApply source_red_safe_implies.
      iIntros ([l_s ?]); simplify_eq.
    }
    destruct x2 as [[| | | |ly_s |]| | |] => //.
    iReIntro "H7".

    rename x2 into l_t. rename x5 into l_s. rename x into qx.
    rename x7 into vx_t. rename x6 into vx_s.
    rename x8 into lr_s. rename x0 into li_t. rename x4 into li_s.
    rename x3 into π.

    (* case analysis: do [x] and [y] alias? *)
    destruct (decide (ly_s = l_s)) as [-> | Hneq].
    - (* they do alias, so we do not need to exploit a second time. *)
      iPoseProof (heapbij_loc_inj with "H6 H7") as "->".
      iSteps. rename x0 into lr_t. rename x into li_t.
      set inv := (
        ∃ (z_i : Z) (vr_t vr_s : val),
        l_t ↦t{#qx} vx_t ∗
        l_s ↦s{#qx} vx_s ∗
        li_s ↦s #z_i ∗
        li_t ↦t #z_i ∗
        lr_t ↦t vr_t ∗
        lr_s ↦s vr_s ∗
        val_rel vr_t vr_s ∗
        na_locs π (<[l_s:=(l_t, NaRead qx)]> ∅) ∗
        †li_s…s 1 ∗ †lr_s…s 1 ∗ †li_t…t 1 ∗ †lr_t…t 1)%I.

      sim_bind (While _ _) (While _ _).
      iApply (sim_while_while inv with "[-]"); iSteps.
      + rewrite bool_decide_false //.
        iSteps. iApply sim_expr_base. iSteps.
      + rewrite bool_decide_true //. do 4 iStep.
        iStepDebug. iLeft. iSteps.
    - (* the proof of this case is very similar to the one above, except that we carry around
        the additional locations in the invariant and have to release both in the end. *)
      iApply (sim_bij_exploit_load with "H7 H12"); [| |]. {
        intros.
        safe_reach_fill (While _ _)%E => /=.
        apply: safe_reach_pure; [eapply pure_while | done | ].
        safe_reach_fill (Load _ _)%E.
        apply: safe_reach_safe_implies => ?.
        apply: safe_reach_refl. apply: post_in_ectx_intro. naive_solver.
      }
      { rewrite lookup_insert_ne; last done. apply lookup_empty. }
      iSteps. rename li_t into lx_t. rename l_s into lx_s.
      rename x into qy. rename l_t into ly_t. rename x0 into vy_t.
      rename x1 into vy_s. rename x2 into li_t. rename x3 into lr_t.

      set inv := (
        ∃ (z_i : Z) (vr_t vr_s : val),
        lx_t ↦t{#qx} vx_t ∗
        lx_s ↦s{#qx} vx_s ∗
        ly_t ↦t{#qy} vy_t ∗
        ly_s ↦s{#qy} vy_s ∗
        li_s ↦s #z_i ∗
        li_t ↦t #z_i ∗
        lr_t ↦t vr_t ∗
        lr_s ↦s vr_s ∗
        val_rel vr_t vr_s ∗
        na_locs π (<[ly_s:=(ly_t, NaRead qy)]> $ <[lx_s:=(lx_t, NaRead qx)]> ∅) ∗
        †li_s…s 1 ∗ †lr_s…s 1 ∗ †li_t…t 1 ∗ †lr_t…t 1)%I.

      sim_bind (While _ _) (While _ _).
      iApply (sim_while_while inv with "[-] []"); iSteps.
      + destruct (bool_decide (#x = vy_t)) eqn:Heq_vy_t.
        { (* contradiction *)
          apply bool_decide_eq_true_1 in Heq_vy_t. subst vy_t.
          val_discr_target "H17". done.
        }
        iSteps. iApply sim_expr_base. iSteps.
      + rewrite bool_decide_true //. 
        do 4 iStep. iStepDebug. iLeft. iSteps. rewrite delete_notin; last first.
        { rewrite lookup_insert_ne //. }
        iSteps.
    Qed.

  (** This optimization hoists the read of "n" out of the loop. *)
  Definition hoist_load_s : expr :=
     let: "r" := ref #0  in
     let: "i" := ref #0  in
     while: !"i" ≠ ! "n" do
       "r" <- !"r" + !"i";;
       "i" <- !"i" + #1
     od;;
     let: "res" := !"r" in
     Free "r";;
     Free "i";;
     "res".

  Definition hoist_load_t : expr :=
     let: "r" := ref #0  in
     let: "i" := ref #0  in
     let: "nval" := ! "n" in
     while: !"i" ≠ "nval" do
       "r" <- !"r" + !"i";;
       "i" <- !"i" + #1
     od;;
     let: "res" := !"r" in
     Free "r";;
     Free "i";;
     "res".


  Lemma hoist_load_sim:
    ⊢ log_rel hoist_load_t hoist_load_s.
  Proof.
    log_rel. iSteps.
    destruct (if x0 is #(LitLoc _) then true else false) eqn:Heq. 2: {
      source_while.
      source_bind (! _)%E. iApply source_red_safe_implies.
      iIntros ([l_s ?]); simplify_eq.
    }
    destruct x0 as [[| | | |n_s |]| | |] => //.
    iReIntro "H4".
    iApply (sim_bij_exploit_load with "H4 H5"); [|done|]. {
      intros.
      safe_reach_fill (While _ _)%E => /=.
      apply: safe_reach_head_step; [ by econstructor|].
      safe_reach_fill (! _)%E => /=.
      apply: safe_reach_safe_implies => ?.
      apply: safe_reach_refl. apply: post_in_ectx_intro. naive_solver.
    }
    iSteps.
    sim_bind (While _ _) (While _ _).
    (* the invariant: we just thread through our ownership, with the invariant that
      [r] and [i] point to the same integer values.
      (retaining this information eases our live in the loop proof significantly)
    *)
    rename x0 into n_t. rename x6 into v_t. rename x7 into v_s. rename x into q.
    rename x2 into li_s. rename x3 into lr_s. rename x4 into li_t. rename x5 into lr_t.
    rename x1 into π.
    set (inv := (
      ∃ z_i z_r : Z,
      †lr_s…s 1 ∗ †li_s…s 1 ∗ †lr_t…t 1 ∗ †li_t…t 1 ∗
      lr_s ↦s #z_r ∗ li_s ↦s #z_i ∗ lr_t ↦t #z_r ∗ li_t ↦t #z_i ∗
      n_t ↦t{#q} v_t ∗ n_s ↦s{#q} v_s ∗ na_locs π (<[n_s:=(n_t, NaRead q)]> ∅))%I).
    iApply (sim_while_while inv with "[-]"); iSteps.
    - destruct (bool_decide (#x0 = v_t)) eqn:Heq_v_t.
      { (* contradiction *)
        apply bool_decide_eq_true_1 in Heq_v_t. subst v_t.
        val_discr_target "H15". done.
      }
      iSteps. iApply sim_expr_base. iSteps.
    - rewrite bool_decide_true //. do 4 iStep. iStepDebug. iLeft. iSteps.
  Qed.

  (** This optimization hoists the read of "m" out of the loop. We
  need to unroll the loop once because the loop condition can be an
  arbitrary expression [e] that does not write to the heap. *)
  Definition hoist_load_unknown_s (e : expr) : expr :=
     let: "r" := ref #0  in
     let: "i" := ref #0  in
     while: e do
       "r" <- !"r" + !"m";;
       "i" <- !"i" + #1
     od;;
     Free "i";;
     let: "res" := !"r" in
     Free "r";;
     "res".

  Definition hoist_load_unknown_t (e : expr) : expr :=
    let: "i" := ref #0 in
    if: e then
      let: "mval" := !"m" in
      let: "r" := ref "mval" in
      "i" <- !"i" + #1;;
      while: e do
        "r" <- !"r" + "mval";;
        "i" <- !"i" + #1
      od;;
      Free "i";;
      let: "res" := !"r" in
      Free "r";;
      "res"
    else
      Free "i";;
      #0.

  Lemma hoist_load_unknown_sim e:
    free_vars e ⊆ list_to_set ["n"; "i"] →
    gen_expr_wf readonly_wf e →
    ⊢ log_rel (hoist_load_unknown_t e) (hoist_load_unknown_s e).
  Proof.
    move => He ?. log_rel. iSteps. 
    iApply sim_update_si. iApply (sim_bij_freeable_ne_val with "H6 H12"). iIntros (Hne3) "Hfi_s".
    iApply sim_update_si. iApply (sim_bij_freeable_ne_val with "H6 H10"). iIntros (Hne1) "Hfr_s".
    iApply (sim_bij_insert with "H14 Hfi_s H13 H11 []"); [done..|]. iIntros "#Hbiji".
    source_while. to_sim.
    sim_bind (subst_map _ _) (subst_map _ _).
    iApply (sim_refl with "[] [H8]");
      [compute_done | etrans; [eassumption|compute_done]
       | apply: readonly_log_rel_structural [] ∅ | done | | |]. {
        rewrite !dom_insert_L. iApply big_sepS_intro. iIntros "!#" (x7 Hin).
        rewrite map_lookup_zip_with.
        destruct (decide (x7 = "n")); [| destruct (decide (x7 = "i")); [|exfalso; set_solver]]; by simplify_map_eq.
    }
    { iFrame. unfold mapsto_list, na_locs_in_mapsto_list. iSplit; [|done]. iPureIntro. set_solver. }
    iSteps. destruct x9; iSteps.
    { rewrite lookup_insert_ne //. contradict Hne3; subst => //. }
    sim_bind (While _ _) (While _ _).
    rename x0 into l_t. rename x7 into l_s.
    rename x into q. rename x11 into m. rename x3 into π.
    rewrite insert_insert.
    assert (x5 ≠ l_s). { contradict Hne3. by subst. }
    assert (x4 ≠ l_s); first (contradict Hne1; by subst).
    iApply (sim_bij_insert with "H18 Hfr_s H12 H17 []"); [done..|]. iStep.
    iPoseProof (sim_bij_release with "Hbiji H20") as "H". { rewrite lookup_insert //. }
    iApply ("H" with "H15 H21 []"); first iSteps. iStep.
    rewrite delete_insert; last rewrite lookup_insert_ne //.
    iApply (sim_while_while (l_t ↦t{#q} #m ∗ l_s ↦s{#q} #m ∗ na_locs π (<[l_s:=(l_t, NaRead q)]> ∅))%I with "[-]").
    { iSteps. }

    iSteps.
    sim_bind (subst_map _ _) (subst_map _ _).
    iApply (sim_refl with "[] [-]");
      [compute_done | etrans; [eassumption|compute_done]
       | apply (readonly_log_rel_structural [(l_t, l_s, #m ,#m , q)]) | done | | |]. {
      rewrite !dom_insert_L. iApply big_sepS_intro. iIntros "!#" (x Hin).
      rewrite map_lookup_zip_with.
      destruct (decide (x = "n")); [| destruct (decide (x = "i")); [|exfalso; set_solver]]; by simplify_map_eq.
    } {
      iFrame "∗ H13". iSplit; [| by eauto ]. iPureIntro.
      move => ??? /lookup_insert_Some[[??]|[??]]; simplify_eq. by eexists 0, _, _.
    }
    iSteps. iReIntro "H20". destruct x3.
    - do 36 iStep. rewrite insert_insert.
      iPoseProof (sim_bij_release with "H17 H24") as "H". { rewrite lookup_insert //. }
      iApply ("H" with "H18 H25 []"); first iSteps. iStep.
      rewrite delete_insert; last rewrite lookup_insert_ne //.
      iSteps. { rewrite lookup_insert_ne //. }
      rewrite insert_insert.
      iPoseProof (sim_bij_release with "Hbiji H26") as "H". { rewrite lookup_insert //. }
      iApply ("H" with "H24 H27 []"); first iSteps. iStep. rewrite delete_insert; last rewrite lookup_insert_ne //.
      iApply sim_expr_base. iSteps.
    - do 7 iStep. iStepDebug. iLeft. iSteps. { rewrite lookup_insert_ne //. }
      rewrite delete_insert; last rewrite lookup_insert_ne //.
      iSteps. Unshelve.
      rewrite lookup_insert_ne //.
  Qed.

End data_race.

Section closed.
  (** Obtain a closed proof of [ctx_ref]. *)
  Lemma hoist_load_ctx :
    ctx_ref hoist_load_t hoist_load_s.
  Proof.
    intros ??.
    set Σ := #[naΣ].
    apply (log_rel_adequacy Σ)=>?.
    by apply hoist_load_sim.
  Qed.

  (** Obtain a closed proof of [ctx_ref]. *)
  Lemma hoist_load_both_ctx :
    ctx_ref hoist_load_both_t hoist_load_both_s.
  Proof.
    intros ??.
    set Σ := #[naΣ].
    apply (log_rel_adequacy Σ)=>?.
    by apply hoist_load_both_sim.
  Qed.

  (** Obtain a closed proof of [ctx_ref]. *)
  Lemma hoist_load_unknown_ctx e :
    free_vars e ⊆ list_to_set ["n"; "i"] →
    gen_expr_wf readonly_wf e →
    ctx_ref (hoist_load_unknown_t e) (hoist_load_unknown_s e).
  Proof.
    intros ??.
    set Σ := #[naΣ].
    apply (log_rel_adequacy Σ)=>?.
    by apply hoist_load_unknown_sim.
  Qed.
End closed.
